# -*- coding: utf-8 -*-
"""
Created on Fri Jan 24 17:47:46 2025

@author: dvezinet
"""


import os
import warnings


import numpy as np
import datastock as ds


from ._utils_types import _DTYPES
from . import _class01_load_from_eqdsk as _eqdsk
from . import _class01_load_from_meq as _meq


# ########################################################
# ########################################################
#               EQDSK
# ########################################################


def main(
    dpfe=None,
    returnas=None,
    # keys
    kmesh=None,
    # user-defined dunits
    dunits=None,
    # group naming
    func_key_groups=None,
    # sorting
    sort_vs=None,
    # derived
    add_rhopn=None,
    add_BRZ=None,
    # optipns
    verb=None,
    strict=None,
    explore=None,
):
    """ load multiple eqdsk equilibria files and concatenate them

    If provided, vector t is used as the time vector
    Otherwise t is just a range

    Parameters
    ----------
    dpfe : str of dict of {path: [patterns]}
        DESCRIPTION. The default is None.
    returnas : dict, True or Collection
        DESCRIPTION. The default is None.
    kmesh : str
        DESCRIPTION. The default is None.
    t : sequence, optional
        DESCRIPTION. The default is None.

    Raises
    ------
    error
        DESCRIPTION.
    Exception
        DESCRIPTION.

    Returns
    -------
    out : dict or Collection
        DESCRIPTION.

    """

    # --------------------
    # check inputs
    # --------------------

    (
        lpfe,
        eqtype, deqtype,
        returnas, coll,
        kmesh,
        func_key_groups,
        sort_vs,
        add_rhopn,
        add_BRZ,
        verb,
        strict,
        explore,
    ) = check_inputs(
        dpfe=dpfe,
        returnas=returnas,
        kmesh=kmesh,
        # user-defined dunits
        dunits=dunits,
        # group naming
        func_key_groups=func_key_groups,
        # sorting
        sort_vs=sort_vs,
        # derived
        add_rhopn=add_rhopn,
        add_BRZ=add_BRZ,
        # options
        verb=verb,
        strict=strict,
        explore=explore,
    )

    # ----------------
    # double-check shapes
    # ----------------

    lpfe, dout, dgroups = _check_shapes(
        lpfe=lpfe,
        deqtype=deqtype,
        eqtype=eqtype,
        # group keys
        func_key_groups=func_key_groups,
        # options
        verb=verb,
        strict=strict,
        explore=explore,
    )

    # ----------------
    # load and extract
    # ----------------

    dfail = {}
    dout_group = {}
    for ig, (kg, vg) in enumerate(dgroups.items()):

        dref, ddata, dmesh = _initialize(
            dgroups=dgroups,
            kgroup=kg,
            kmesh=kmesh,
            deqtype=deqtype,
            eqtype=eqtype,
            lkeys=vg['lkeys'],
            ltypes=vg['ltypes'],
            dout=dout[vg['pfe'][0]]
        )

        for ip, pfe in enumerate(vg['pfe']):

            # -----------------------
            # check grid mRZ and mrho

            c0, msg = _check_dmesh(dmesh, deqtype, pfe, dout, dmesh)
            if c0 is False:
                dfail[pfe] = msg
                continue

            # ---------------
            # fill all fields

            for ia, (katt, typ) in enumerate(zip(vg['lkeys'], vg['ltypes'])):

                if katt not in deqtype['dunits'].keys():
                    continue

                if typ == 'str':
                    ddata[katt]['data'].append(dout[pfe][katt])

                elif typ in ['scalar', 'bool']:
                    ddata[katt]['data'][ip] = dout[pfe][katt]

                else:
                    sli = [slice(None) for ss in dout[pfe][katt].shape]
                    sli = (ip,) + tuple(sli)
                    if deqtype['dunits'][katt].get('transpose', False) is True:
                        ddata[katt]['data'][sli] = dout[pfe][katt].T
                    else:
                        ddata[katt]['data'][sli] = dout[pfe][katt]

            # ---------------
            # derived rhopn

            if add_rhopn is True:
                ddata['rhopn'] = _add_rhopn(ddata, dunits=deqtype['dunits'])

            # ---------------
            # derived BRZ

            if add_BRZ is True:
                # ddata['BR'], ddata['BZ'] = _add_BRZ()
                pass

        # ----------------
        # str list to array

        for kk, vv in ddata.items():
            if isinstance(vv['data'], list):
                ddata[kk]['data'] = np.array(vv['data'])

        # ----------------
        # optional sorting

        if sort_vs is not None:

            lok_att = [kk for kk in ddata.keys() if len(ddata[kk]['ref']) == 1]
            lok_key = [ddata[kk]['key'] for kk in lok_att]
            if sort_vs in lok_key:
                sort_vs = [
                    kk for kk in lok_att if ddata[kk]['key'] == sort_vs
                ][0]
            if sort_vs in lok_att:
                inds = np.argsort(ddata[sort_vs]['data'])
                for kk, vv in ddata.items():
                    sli = (inds,) + tuple([slice(None) for ss in vv['data'].shape[1:]])
                    ddata[kk]['data'] = vv['data'][sli]

        # -----------------------
        # store as group-specific

        dout_group[vg['key']] = {
            'dref': dref,
            'dmesh': dmesh,
            'ddata': ddata,
        }

    # ----------------------
    # raise error if needed
    # ---------------------

    if len(dfail) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "The following files could not be loaded:\n"
            + "\n".join(lstr)
        )
        if strict is True:
            raise Exception(msg)
        else:
            warnings.warn(msg)

    # -------------------
    # build output
    # -------------------

    if returnas is dict:
        return dout_group

    else:
        _to_Collection(
            coll=coll,
            dout_group=dout_group,
        )


# ########################################################
# ########################################################
#               check
# ########################################################


def check_inputs(
    dpfe=None,
    returnas=None,
    # keys
    kmesh=None,
    # user-defined dunits
    dunits=None,
    # group naming
    func_key_groups=None,
    # sorting
    sort_vs=None,
    # derived
    add_rhopn=None,
    add_BRZ=None,
    # optipns
    verb=None,
    strict=None,
    explore=None,
):

    # --------------------
    # check pfe
    # --------------------

    lpfe = ds.get_files(
        dpfe=dpfe,
        returnas=list,
        strict=strict,
    )

    # ----------------------------
    # check file type consistency

    lext = sorted(set([pfe.split('.')[-1] for pfe in lpfe]))
    if len(lext) > 1:
        msg = (
            "Provided dpfe led to loading files of different types!\n"
            f"Indentified extensions: {lext}\n\n"
            f"Provided dpfe:\n{dpfe}\n\n"
            f"Resulting files:\n"
            + "\n".join([f"\t- {pfe}" for pfe in lpfe])
        )
        raise Exception(msg)
    ext = lext[0].lower()

    # --------------------
    # dunits
    # --------------------

    if dunits is not None:
        ls = ['key', 'units', 'ref', 'transpose']
        c0 = (
            isinstance(dunits, dict)
            and all([
                isinstance(k0, str)
                and isinstance(v0, dict)
                and all([ss in ls for ss in v0.keys()])
                for k0, v0 in dunits.items()
            ])
        )
        if not c0:
            msg = (
                "Arg 'dunits', if provided must be a dict with subdicts:\n"
                "{"
                "\t...,\n"
                "\t'Brx': {\n"
                "\t\t'key': 'BR',\n"
                "\t\t'units': 'T',\n"
                "\t\t'ref': ('neq', 'mRZ'),\n"
                "\t\t'transpose': True,\n"
                "\t},\n"
                "\t...,\n"
                "}\n\n"
                "If not provided, loaded from one of:\n"
                "\t- tofu/data/_class01_load_from_eqdsk.py\n"
                "\t- tofu/data/_class01_load_from_meq.py\n\n"
                "Use explore=True to investigate files content!\n"
                f"\nProvided:\n{dunits}"
            )
            raise Exception(msg)

    else:
        if ext == 'eqdsk':
            dunits = _eqdsk._DUNITS
        else:
            dunits = _meq._DUNITS

    # --------------------
    # get load_pfe and dunits
    # --------------------

    if ext == 'eqdsk':
        eqtype = 'eqdsk'
        deqtype = {
            'load_pfe': _eqdsk.get_load_pfe(),
            'dunits': dunits,
            'extract_grid': _eqdsk._extract_grid,
            'extra_keys': _eqdsk._EXTRA_KEYS,
        }

    elif ext == 'mat':
        eqtype = 'meq'
        deqtype = {
            'load_pfe': _meq.get_load_pfe(),
            'dunits': dunits,
            'extract_grid': _meq._extract_grid,
            'extra_keys': [],
        }

    else:
        msg = (
            "Non-reckognized file extension!\n"
            f"\t- allowed: ['eqdsk', 'mat']\n"
            f"\t- Provided: {ext}"
        )
        raise Exception(msg)

    # --------------------
    # check returnas
    # --------------------

    if returnas is None:
        returnas = True

    coll = None
    from ._class10_Inversion import Inversion as Collection
    if returnas is True:
        coll = Collection()

    elif issubclass(returnas.__class__, Collection):
        coll = returnas

    elif returnas is not dict:
        msg = (
            "returnas must be either:\n"
            "\t- dict: return mesh in ddata and dref dict\n"
            "\t- True: return mesh in new Collection instance\n"
            "\t- Collection instance: add mesh 2d rect to it, named kmesh"
        )
        raise Exception(msg)

    # -------------------
    # kmesh
    # -------------------

    if coll is not None:

        if kmesh is None:
            kmesh = {}
        if isinstance(kmesh, str):
            kmesh = {'mRZ': kmesh}

        wm = coll._which_mesh
        for km in ['mRZ', 'mrhotn']:
            kmesh[km] = ds._generic_check._obj_key(
                d0=coll.dobj.get(wm, {}),
                short='m',
                key=kmesh.get(km, km),
                ndigits=2,
            )

    # ---------------
    # group naming
    # ---------------

    if func_key_groups is None:
        func_key_groups = lambda pfe, i0: f"{eqtype}{i0}"

    else:
        try:
            func_key_groups('', 0)
        except Exception as err:
            msg = (
                "If provided, func_key_groups must be a callable of 2 args:\n"
                "\t- pfe: str to file\n"
                "\t- ii: int\n"
                "Provided:\n{func_key_groups}\n"
            )
            raise Exception(msg) from err

    # ---------------
    # sort_vs
    # ---------------

    if sort_vs is not None:
        sort_vs = ds._generic_check._check_var(
            sort_vs, 'sort_vs',
            types=str,
        )

    # ---------------
    # add_rhopn
    # ---------------

    add_rhopn = ds._generic_check._check_var(
        add_rhopn, 'add_rhopn',
        types=bool,
        default=True,
    )

    # ---------------
    # add_BRZ
    # ---------------

    add_BRZ = ds._generic_check._check_var(
        add_BRZ, 'add_BRZ',
        types=bool,
        default=True,
    )

    # ---------------
    # verb
    # ---------------

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=False,
    )

    # ---------------
    # strict
    # ---------------

    strict = ds._generic_check._check_var(
        strict, 'strict',
        types=bool,
        default=False,
    )

    # ---------------
    # explore
    # ---------------

    explore = ds._generic_check._check_var(
        explore, 'explore',
        types=bool,
        default=False,
    )

    return (
        lpfe,
        eqtype, deqtype,
        returnas, coll,
        kmesh,
        func_key_groups,
        sort_vs,
        add_rhopn,
        add_BRZ,
        verb,
        strict,
        explore,
    )


# ########################################################
# ########################################################
#               check shapes
# ########################################################


def _check_shapes(
    lpfe=None,
    deqtype=None,
    eqtype=None,
    # group keys
    func_key_groups=None,
    # options
    verb=None,
    strict=None,
    explore=None,
):

    # ------------------
    # prepare
    # -----------------

    # loop on all files
    dfail = {}
    dgroups = {}
    dout = {}

    # ------------------
    # loop on files
    # -----------------

    i0 = 0
    for ii, pfe in enumerate(lpfe):

        # --------------
        # open and load

        try:
            dpfe = deqtype['load_pfe'](pfe)
        except Exception as err:
            if strict is True:
                raise err
            dfail[pfe] = err
            continue

        # --------------------
        # standardize content

        dpfe = _standardize(dpfe, pfe)

        # -----------------
        # sorted attributes and types

        # attributes
        lk = sorted(dpfe.keys())

        # types
        ltypes = [
            dpfe[kk].shape if isinstance(dpfe[kk], np.ndarray)
            else (
                'str' if isinstance(dpfe[kk], _DTYPES['str'])
                else (
                    'scalar' if isinstance(dpfe[kk], _DTYPES['scalar'])
                    else 'bool'
                )
            )
            for kk in lk
        ]

        # -------------
        # explore (optional)

        if explore is True:
            _explore(lk, ltypes, dpfe, pfe, deqtype['dunits'])

        # ---------------------
        # unique field_type key

        # reduce to only loaded fron dunits
        lik = [
            ik for ik, kk in enumerate(lk)
            if kk in deqtype['dunits'].keys()
            or kk in deqtype['extra_keys']
        ]
        lk = [lk[ik] for ik in lik]
        ltypes = [ltypes[ik] for ik in lik]

        # field_types
        field_types = "_".join([
            f"({kk}, {tt})" for kk, tt in zip(lk, ltypes)
        ])

        # -----------
        # groups

        if field_types not in dgroups.keys():
            kgroup = func_key_groups(pfe, i0)
            dgroups[field_types] = {
                'key': kgroup,
                'lkeys': lk,
                'ltypes': ltypes,
                'pfe': [pfe],
            }
            i0 += 1

        else:
            dgroups[field_types]['pfe'].append(pfe)

        # ----------
        # initialize

        dout[pfe] = {
            'group': kgroup,
            **{kk: dpfe[kk] for kk in lk},
        }

    # ----------------
    # dfail
    # ----------------

    if len(dfail) > 0:

        lfail = sorted(dfail.keys())
        lpath = [os.path.split(pfe) for pfe in lfail]
        pathu = sorted(set([pp[0] for pp in lpath]))
        dpathu = {
            path: [pp[1] for pp in lpath if pp[0] == path]
            for path in pathu
        }

        lstr = [
            f"path = {path}:\n"
            + "\n".join([
                f"\t- {pp}: {str(dfail[os.path.join(path, pp)])}"
                for pp in dpathu[path]
            ])
            for path in pathu
        ]

        msg = (
            "The following files could not be loaded:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # -------------
    # verb
    # -------------

    if verb is True:

        lstr = [
            f"\n{eqtype} group '{kg}':\n"
            + "\n".join([
                f"\t- {kk}: {typ}"
                for kk, typ in zip(vg['lkeys'], vg['ltypes'])
            ])
            for kg, vg in dgroups.items()
        ]
        msg = (
            "\n-----------------------\nDetected content of files:\n"
            + "\n".join(lstr)
        )
        print(msg)

    # -------------
    # adjust lpfe
    # -------------

    lpfe = [pfe for pfe in lpfe if pfe not in dfail.keys()]

    return lpfe, dout, dgroups


# ########################################################
# ########################################################
#               Standardize
# ########################################################


def _standardize(dpfe, pfe):

    dout = {}
    for k0, v0 in dpfe.items():

        if isinstance(v0, np.ndarray):
            if v0.size > 1:
                dout[k0] = np.squeeze(v0)
            elif v0.size == 1:
                dout[k0] = _adjust_type(np.squeeze(v0).tolist(), k0, pfe)

        else:
            dout[k0] = _adjust_type(v0, k0, pfe)

    return dout


def _adjust_type(val, key, pfe):

    if isinstance(val, _DTYPES['int']):
        val = int(val)

    elif isinstance(val, _DTYPES['float']):
        val = float(val)

    elif isinstance(val, _DTYPES['bool']):
        val = bool(val)

    elif isinstance(val, _DTYPES['str']):
        val = str(val)

    else:
        msg = (
            "Unreckognized variable type in eqilibrium file!\n"
            f"\t- pfe: {pfe}\n"
            f"\t- variable: {key}\n"
            f"\t- type: {type(val)}"
            f"\t- value: {val}\n"
        )
        raise Exception(msg)

    return val


def _explore(lk, ltypes, dpfe, pfe, dunits):

    lstr = []
    for ii, (kk, tt) in enumerate(zip(lk, ltypes)):
        if tt in ['scalar', 'bool']:
            stri = f"\t- {ii}/{len(lk)-1}\t{kk}: {dpfe[kk]}"
        elif isinstance(tt, tuple):
            if tt == (2,):
                stri = f"\t- {ii}/{len(lk)-1}\t{kk}: {dpfe[kk]}"
            else:
                stri = stri = f"\t- {ii}/{len(lk)-1}\t{kk}: {tt}"
        lstr.append(stri)

    msg = (
        f"\n\nExploring content of file: {pfe}\n"
        + "\n".join(lstr)
    )
    print(msg)

    return


# ########################################################
# ########################################################
#               Initialize
# ########################################################


def _initialize(
    dgroups=None,
    kgroup=None,
    kmesh=None,
    deqtype=None,
    eqtype=None,
    lkeys=None,
    ltypes=None,
    dout=None,
):

    # ------------
    # nb of pfe
    # ------------

    npfe = len(dgroups[kgroup]['pfe'])

    krpfe = "neq"
    dref = {
        'neq': {
            'key': krpfe,
            'size': npfe,
        },
    }

    kipfe = 'ieq'
    ddata= {
        kipfe: {
            'key': kipfe,
            'data': np.arange(0, npfe),
            'ref': (krpfe,),
            'dim': 'index',
            'units': None,
        },
        'pfe': {
            'key': 'pfe',
            'data': np.array(dgroups[kgroup]['pfe']),
            'ref': (krpfe,),
            'dim': 'index',
            'units': None,
        },
    }

    # ------------
    # R, Z grid
    # ------------

    dmesh = deqtype['extract_grid'](dout, kmesh)

    # ------------
    # lim grid
    # ------------

    if eqtype == 'eqdsk':
        nsep = dout['nbdry']
    else:
        nsep = dout['rS'].size

    krnsep = 'nsep'
    dref['nsep'] = {
        'key': krnsep,
        'size': nsep,
    }

    # ------------
    #
    # ------------

    for ii, (katt, typ) in enumerate(zip(lkeys, ltypes)):

        if katt not in deqtype['dunits'].keys():
            continue

        # init
        if typ == 'str':
            init = []
        elif typ == 'scalar':
            init = np.full((npfe,), np.nan)
        elif typ == 'bool':
            init = np.zeros((npfe,), dtype=bool)
        else:
            if deqtype['dunits'][katt].get('transpose', False) is True:
                typ = typ[::-1]
            shape = (npfe,) + typ
            init = np.full(shape, np.nan)

        # ref
        ref = tuple([
            dref[rr]['key'] if rr in dref.keys() else f"{dmesh[rr]['key']}_bs1"
            for rr in deqtype['dunits'][katt]['ref']
        ])

        # data
        key = deqtype['dunits'][katt].get('key', katt)
        ddata[katt] = {
            'key': key,
            'data': init,
            'units': deqtype['dunits'][katt]['units'],
            'ref': ref,
        }

    return dref, ddata, dmesh


# ########################################################
# ########################################################
#               check dmesh
# ########################################################


def _check_dmesh(dmesh, deqtype, pfe, dout, kmesh):

    dm = deqtype['extract_grid'](dout[pfe], kmesh)

    msg = None
    c0 = True
    for km, vm in dm.items():
        lk = [kk for kk in vm.keys() if 'knots' in kk]
        for kk in lk:
            if not np.allclose(vm[kk], dmesh[km][kk]):
                msg = f"mesh '{km}' has different knots '{kk}'"
                c0 = False
                continue
        if c0 is False:
            break

    return c0, msg


# ########################################################
# ########################################################
#               Derived
# ########################################################


def _add_rhopn(ddata=None, dunits=None):

    kpsi = [kk for kk, vv in dunits.items() if vv['key'] == 'psi'][0]
    kpsi0 = [kk for kk, vv in dunits.items() if vv['key'] == 'psi_axis'][0]
    kpsi1 = [kk for kk, vv in dunits.items() if vv['key'] == 'psi_sep'][0]

    psi0 = ddata[kpsi0]['data']
    psi1 = ddata[kpsi1]['data']
    psi = ddata[kpsi]['data']

    rhopn = (
        (psi0[:, None, None] - psi)
        / (psi0[:, None, None] -psi1[:, None, None])
    )

    return {
        'key': 'rhopn',
        'data': rhopn,
        'units': None,
        'ref': ddata[kpsi]['ref'],
    }


# ########################################################
# ########################################################
#               to Collection
# ########################################################


def _to_Collection(
    coll=None,
    dout_group=None,
):

    # ---------------
    # loop on groups
    # ---------------

    for kg, vg in dout_group.items():

        # ----------
        # add refs

        for kr, vr in vg['dref'].items():
            vr['key'] = f"{kg}_{vr['key']}"
            coll.add_ref(**vr)

        # ----------
        # add mesh

        for kr, vr in vg['dmesh'].items():
            vr['key'] = f"{kg}_{vr['key']}"
            coll.add_mesh_2d_rect(**vr)

        # ----------
        # add data

        for kr, vr in vg['ddata'].items():

            vr['ref'] = tuple([f"{kg}_{rr}" for rr in vr['ref']])
            vr['key'] = f"{kg}_{vr['key']}"
            coll.add_data(**vr)

    return




import copy
import warnings


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines


import Polygon as plg
import datastock as ds


from . import _comp_solidangles


__all__ = ['compute_etendue']


# #############################################################################
# #############################################################################
#                    Main routines
# #############################################################################


def compute_etendue(
    det=None,
    aperture=None,
    analytical=None,
    numerical=None,
    check=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    verb=None,
    plot=None,
):
    """ Only works for a set of detectors associated to a single aperture

    Typical use:
        - pinhole cameras

    """

    # -------------
    # check inputs

    (
        det, ldeti, aperture,
        analytical, numerical,
        res, margin_par, margin_perp,
        check, verb, plot,
    ) = _compute_etendue_check(
        det=det,
        aperture=aperture,
        analytical=analytical,
        numerical=numerical,
        res=res,
        margin_par=margin_par,
        margin_perp=margin_perp,
        check=check,
        verb=verb,
        plot=plot,
    )

    # ----------
    # prepare

    (
        det_area, ap_area, distances,
        los_x, los_y, los_z,
        cos_los_det, cos_los_ap,
        solid_angles, res, pix_ap,
    ) = _compute_etendue_prepare(
        ldeti=ldeti,
        aperture=aperture,
        res=res,
    )

    shape = distances.shape

    # --------------------
    # compute analytically

    if analytical is True:
        etend0 = np.full(tuple(np.r_[3, shape]), np.nan)

        # 0th order
        etend0[0, :] = ap_area * det_area / distances**2

        # 1st order
        etend0[1, :] = (
            cos_los_ap * ap_area
            * cos_los_det * det_area / distances**2
        )

        # 2nd order
        etend0[2, :] = cos_los_ap * ap_area * solid_angles

    else:
        etend0 = None

    # --------------------
    # compute numerically

    if numerical is True:
        etend1 = _compute_etendue_numerical(
            ldeti=ldeti,
            aperture=aperture,
            pix_ap=pix_ap,
            res=res,
            los_x=los_x,
            los_y=los_y,
            los_z=los_z,
            margin_par=margin_par,
            margin_perp=margin_perp,
            check=check,
            verb=verb,
        )

    else:
        etend1 = None

    # --------------------
    # optional plotting

    if plot is True:
        dax = _plot_etendues(
            etend0=etend0,
            etend1=etend1,
            res=res,
        )

    # --------
    # reshape

    sh0 = det['cents_x'].shape
    if sh0 == ():
        sh0 = (1,)

    # etend0
    if etend0 is not None and etend0.shape[1:] != sh0:
        etend0 = etend0.reshape(tuple(np.r_[etend0.shape[0], sh0]))

    # etend1
    if etend1 is not None and etend1.shape[1:] != sh0:
        etend1 = etend1.reshape(tuple(np.r_[res.size, sh0]))

    # los
    if los_x.shape != sh0:
        los_x = los_x.reshape(sh0)
        los_y = los_y.reshape(sh0)
        los_z = los_z.reshape(sh0)

    # --------------------
    # return

    dout = {
        'analytical': etend0,
        'numerical': etend1,
        'res': res,
        'los_x': los_x,
        'los_y': los_y,
        'los_z': los_z,
    }

    return dout


# #############################################################################
# #############################################################################
#                   input checking routine
# #############################################################################


def _compute_etendue_check(
    det=None,
    aperture=None,
    analytical=None,
    numerical=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    check=None,
    verb=None,
    plot=None,
):
    """ Check conformity of inputs

    """

    # -----------
    # det

    # check keys
    lk = [
        'cents_x', 'cents_y', 'cents_z',
        'nin_x', 'nin_y', 'nin_z',
        'e0_x', 'e0_y', 'e0_z',
        'e1_x', 'e1_y', 'e1_z',
        'outline_x0', 'outline_x1',
    ]

    c0 = (
        isinstance(det, dict)
        and all([kk in det.keys() for kk in lk])
    )
    if not c0:
        lstr = [f"\t- {k0}" for k0 in lk]
        msg = (
            "Arg det must be a dict with the following keys:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # check values
    for k0 in lk:
        if isinstance(det[k0], (list, tuple)):
            det[k0] = np.atleast_1d(det[k0]).ravel()

        c0 = isinstance(det[k0], np.ndarray) or np.isscalar(det[k0])
        if not c0:
            msg = f"Arg det['{k0}'] must be a np.ndarray or scalar"
            raise Exception(msg)

        if k0 in ['outline_x0', 'outline_x1'] and det[k0].ndim > 1:
            msg = "Arg det['outline_x0'] and det['outline_x1'] must be 1d"
            raise Exception(msg)

    # check shapes
    dshape = {
        0: ['outline_x0', 'outline_x1'],
        1: [
            'nin_x', 'nin_y', 'nin_z',
            'e0_x', 'e0_y', 'e0_z',
            'e1_x', 'e1_y', 'e1_z',
        ],
        2: ['cents_x', 'cents_y', 'cents_z'],
    }
    for k0, v0 in dshape.items():
        if len(set([det[v1].shape for v1 in v0])) > 1:
            lstr = [f"\t- {v1}" for v1 in v0]
            msg = (
                "The following args must share the same shape:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    shaped = det['cents_x'].shape
    if det['cents_x'].shape != det['nin_x'].shape:
        if np.isscalar(det['nin_x']):
            assert det.get('parallel', True) is True
        else:
            msg = (
                "Arg det['nin_x'], det['nin_y'], det['nin_z'] must have "
                "the same shape as det['cents_z']"
            )
            raise Exception(msg)

    # check outline not closed
    if (
        det['outline_x0'][0] == det['outline_x0'][-1]
        and det['outline_x1'][0] == det['outline_x1'][-1]
    ):
        det['outline_x0'] = det['outline_x0'][:-1]
        det['outline_x1'] = det['outline_x1'][:-1]

    # normalization
    norms = np.sqrt(det['nin_x']**2 + det['nin_y']**2 + det['nin_z']**2)
    det['nin_x'] = det['nin_x'] / norms
    det['nin_y'] = det['nin_y'] / norms
    det['nin_z'] = det['nin_z'] / norms

    # -----------
    # ldeti

    lk = [
        k0 for k0 in det.keys()
        if any([k0.endswith(ss) for ss in ['_x', '_y', '_z']])
        and not np.isscalar(det[k0])
    ]
    deti = {k0: det[k0].ravel() for k0 in lk}

    ldeti = [
        {
            k0: deti[k0][ii]
            if k0 in lk else v0
            for k0, v0 in det.items()
        }
        for ii in range(det['cents_x'].size)
    ]

    # -----------
    # aperture

    lk = [
        'poly_x', 'poly_y', 'poly_z',
        'nin', 'e0', 'e1',
    ]

    c0 = (
        isinstance(aperture, dict)
        and all([
            isinstance(k0, str)
            and isinstance(v0, dict)
            and isinstance(v0.get('reflector', False), bool)
            and all([kk in v0.keys() for kk in lk])
            for k0, v0 in aperture.items()
        ])
    )
    if not c0:
        lstr = [f"\t- {k0}" for k0 in lk]
        msg = (
            "Arg aperture must be a dict of sub-dict with keys:\n"
            + "\n".join(lstr)
            + f"\nProvided:\n{aperture}"
        )
        raise Exception(msg)

    # check each case 
    for k0, v0 in aperture.items():

        # check values
        for k1 in lk:
            if isinstance(v0[k1], (list, tuple)):
                aperture[k0][k1] = np.atleast_1d(v0[k1]).ravel().astype(float)

            if not isinstance(aperture[k0][k1], np.ndarray):
                msg = f"Arg aperture['{k0}']['{k1}'] must ba a np.ndarray"
                raise Exception(msg)

            if aperture[k0][k1].ndim > 1:
                msg = f"Arg aperture['{k0}']['{k1}'] must be 1d"
                raise Exception(msg)

            if 'poly_' not in k1 and aperture[k0][k1].shape != (3,):
                msg = f"Arg aperture['{k0}']['{k1}'] must have shape (3,)"
                raise Exception(msg)

        v0 = aperture[k0]

        # check shapes
        dshape = {
            0: ['poly_x', 'poly_y', 'poly_z'],
            1: ['nin', 'e0', 'e1'],
        }
        for k1, v1 in dshape.items():
            if len(set([v0[v2].shape for v2 in v1])) > 1:
                lstr = [f"\t- {v2}" for v2 in v1]
                msg = (
                    "The following args must share the same shape:\n"
                    + "\n".join(lstr)
                )
                raise Exception(msg)

        # check not closed poly
        if (
            v0['poly_x'][0] == v0['poly_x'][-1]
            and v0['poly_y'][0] == v0['poly_y'][-1]
            and v0['poly_z'][0] == v0['poly_z'][-1]
        ):
            v0['poly_x'] = v0['poly_x'][:-1]
            v0['poly_y'] = v0['poly_y'][:-1]
            v0['poly_z'] = v0['poly_z'][:-1]

        # normalization
        norm = np.linalg.norm(v0['nin'])
        v0['nin'] = v0['nin'] / norm

        # derive cents
        if 'cent' not in v0.keys():
            v0['cent'] = np.r_[
                np.mean(v0['poly_x']),
                np.mean(v0['poly_y']),
                np.mean(v0['poly_z']),
            ]

    # -----------
    # analytical

    analytical = ds._generic_check._check_var(
        analytical, 'analytical',
        types=bool,
        default=True,
    )

    # -----------
    # numerical

    numerical = ds._generic_check._check_var(
        numerical, 'numerical',
        types=bool,
        default=True,
    )

    # -----------
    # res

    if res is not None:
        res = np.atleast_1d(res).ravel()

    # -----------
    # margin_par

    margin_par = ds._generic_check._check_var(
        margin_par, 'margin_par',
        types=float,
        default=0.05,
    )

    # -----------
    # margin_perp

    margin_perp = ds._generic_check._check_var(
        margin_perp, 'margin_perp',
        types=float,
        default=0.05,
    )

    # -----------
    # check

    check = ds._generic_check._check_var(
        check, 'check',
        types=bool,
        default=True,
    )

    # -----------
    # verb

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # plot

    if plot is None:
        plot = True
    if not isinstance(plot, bool):
        msg = "Arg plot must be a bool"
        raise Exception(msg)

    return (
        det, ldeti, aperture, analytical, numerical,
        res, margin_par, margin_perp, check, verb, plot,
    )


# #############################################################################
# #############################################################################
#                   preparation routine
# #############################################################################


def _project_poly_on_plane_from_pt(
    pt=None,
    lpoly_x=None,
    lpoly_y=None,
    lpoly_z=None,
    plane_pt=None,
    plane_nin=None,
    plane_e0=None,
    plane_e1=None,
):

    isok = True
    for ii in range(len(lpoly_x)):

        sca0 = (
            (plane_pt[0] - pt[0])*plane_nin[0]
            + (plane_pt[1] - pt[1])*plane_nin[1]
            + (plane_pt[2] - pt[2])*plane_nin[2]
        )

        vx = lpoly_x[ii] - pt[0]
        vy = lpoly_y[ii] - pt[1]
        vz = lpoly_z[ii] - pt[2]

        sca1 = vx*plane_nin[0] + vy*plane_nin[1] + vz*plane_nin[2]

        k = sca0 / sca1

        px = pt[0] + k * vx
        py = pt[1] + k * vy
        pz = pt[2] + k * vz

        p0 = (
            (px - plane_pt[0])*plane_e0[0]
            + (py - plane_pt[1])*plane_e0[1]
            + (pz - plane_pt[2])*plane_e0[2]
        )
        p1 = (
            (px - plane_pt[0])*plane_e1[0]
            + (py - plane_pt[1])*plane_e1[1]
            + (pz - plane_pt[2])*plane_e1[2]
        )

        if ii == 0:
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)

        if p_a.nPoints() < 3:
            isok = False
            break

    if isok is False:
        return None, None, None, None, None, None

    p0, p1 = np.array(p_a.contour(0)).T
    px = plane_pt[0] + p0*plane_e0[0] + p1*plane_e1[0]
    py = plane_pt[1] + p0*plane_e0[1] + p1*plane_e1[1]
    pz = plane_pt[2] + p0*plane_e0[2] + p1*plane_e1[2]

    return p_a, p0, p1, px, py, pz


def _compute_etendue_prepare(
    ldeti=None,
    aperture=None,
    res=None,
):

    # -------------------------
    # intersection of apertures

    kap_ref = list(aperture.keys())[0]
    lpoly_x = [v0['poly_x'] for v0 in aperture.values()]
    lpoly_y = [v0['poly_y'] for v0 in aperture.values()]
    lpoly_z = [v0['poly_z'] for v0 in aperture.values()]

    # prepare data
    nd = len(ldeti)
    ap_area = np.zeros((nd,), dtype=float)
    los_x = np.full((nd,), np.nan)
    los_y = np.full((nd,), np.nan)
    los_z = np.full((nd,), np.nan)
    solid_angles = np.zeros((nd,), dtype=float)
    mindiff = np.full((nd,), np.nan)

    # store projected intersection of apertures (3d), per pix
    # useful later for estimating the plane to be sample (numerical)
    pix_ap = []

    for ii in range(nd):

        # ap
        p_a, p0, p1, px, py, pz = _project_poly_on_plane_from_pt(
            pt=np.r_[
                ldeti[ii]['cents_x'],
                ldeti[ii]['cents_y'],
                ldeti[ii]['cents_z'],
            ],
            lpoly_x=lpoly_x,
            lpoly_y=lpoly_y,
            lpoly_z=lpoly_z,
            plane_pt=aperture[kap_ref]['cent'],
            plane_nin=aperture[kap_ref]['nin'],
            plane_e0=aperture[kap_ref]['e0'],
            plane_e1=aperture[kap_ref]['e1'],
        )

        if p_a is None:
            pix_ap.append(None)
            continue

        else:
            ap_area[ii] = p_a.area()
            ap_cent = (
                aperture[kap_ref]['cent']
                + p_a.center()[0] * aperture[kap_ref]['e0']
                + p_a.center()[1] * aperture[kap_ref]['e1']
            )
            mindiff[ii] = np.sqrt(np.min(np.diff(p0)**2 + np.diff(p1)**2))

            # ----------------------------------
            # los, distances, cosines

            los_x[ii] = ap_cent[0] - ldeti[ii]['cents_x']
            los_y[ii] = ap_cent[1] - ldeti[ii]['cents_y']
            los_z[ii] = ap_cent[2] - ldeti[ii]['cents_z']

            # ------------
            # solid angles

            solid_angles[ii] = _comp_solidangles.calc_solidangle_apertures(
                # observation points
                pts_x=ap_cent[0],
                pts_y=ap_cent[1],
                pts_z=ap_cent[2],
                # polygons
                apertures=None,
                detectors=ldeti[ii],
                # possible obstacles
                config=None,
                # parameters
                visibility=False,
                return_vector=False,
            ).ravel()[0]

            pix_ap.append((px, py, pz))

    # -------------
    # normalize los

    distances = np.sqrt(los_x**2 + los_y**2 + los_z**2)

    los_x = los_x / distances
    los_y = los_y / distances
    los_z = los_z / distances

    # ------
    # angles

    cos_los_det = (
        los_x * ldeti[ii]['nin_x']
        + los_y * ldeti[ii]['nin_y']
        + los_z * ldeti[ii]['nin_z']
    )

    cos_los_ap = (
        los_x * aperture[kap_ref]['nin'][0]
        + los_y * aperture[kap_ref]['nin'][1]
        + los_z * aperture[kap_ref]['nin'][2]
    )

    # -----------
    # surfaces

    # det
    if ldeti[ii].get('pix area') is None:
        det_area = plg.Polygon(np.array([
            ldeti[ii]['outline_x0'],
            ldeti[ii]['outline_x1'],
        ]).T).area()
    else:
        det_area = ldeti[ii]['pix area']

    # -------------------------------------
    # det outline discretization resolution

    if res is None:

        res = min(
            np.sqrt(det_area),
            np.sqrt(np.min(ap_area[ap_area > 0.])),
            np.nanmin(mindiff),
        ) * np.r_[1., 0.5, 0.1]

    iok = np.isfinite(res)
    iok[iok] = res[iok] > 0
    if not np.any(iok):
        res = np.r_[0.001]
    else:
        res = res[iok]

    return (
        det_area, ap_area, distances,
        los_x, los_y, los_z,
        cos_los_det, cos_los_ap, solid_angles, res, pix_ap,
    )


# #############################################################################
# #############################################################################
#           Numerical etendue estimation routine
# #############################################################################


def _compute_etendue_numerical(
    ldeti=None,
    aperture=None,
    pix_ap=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    los_x=None,
    los_y=None,
    los_z=None,
    check=None,
    verb=None,
):

    # shape0 = det['cents_x'].shape
    nd = len(ldeti)

    ap_ind = np.cumsum([v0['poly_x'].size for v0 in aperture.values()][:-1])

    ap_tot_px = np.concatenate(tuple(
        [v0['poly_x'] for v0 in aperture.values()]
    ))
    ap_tot_py = np.concatenate(tuple(
        [v0['poly_y'] for v0 in aperture.values()]
    ))
    ap_tot_pz = np.concatenate(tuple(
        [v0['poly_z'] for v0 in aperture.values()]
    ))

    # ------------------------------
    # Get plane perpendicular to los

    etendue = np.full((res.size, nd), np.nan)
    for ii in range(nd):

        if verb is True:
            msg = f"Numerical etendue for det {ii+1} / {nd}"
            print(msg)

        if np.isnan(los_x[ii]):
            continue

        # get det corners to aperture corners vectors
        out_c_x0 = np.r_[0, ldeti[ii]['outline_x0']]
        out_c_x1 = np.r_[0, ldeti[ii]['outline_x1']]

        # det poly 3d
        det_Px = (
            ldeti[ii]['cents_x']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_x']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_x']
        )
        det_Py = (
            ldeti[ii]['cents_y']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_y']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_y']
        )
        det_Pz = (
            ldeti[ii]['cents_z']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_z']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_z']
        )

        # det to ap vectors
        PA_x = ap_tot_px[:, None] - det_Px[None, :]
        PA_y = ap_tot_py[:, None] - det_Py[None, :]
        PA_z = ap_tot_pz[:, None] - det_Pz[None, :]

        sca1 = PA_x * los_x[ii] + PA_y * los_y[ii] + PA_z * los_z[ii]
        # get length along los
        k_los = (1. + margin_par) * np.max(sca1)

        # get center of plane perpendicular to los
        c_los_x = ldeti[ii]['cents_x'] + k_los * los_x[ii]
        c_los_y = ldeti[ii]['cents_y'] + k_los * los_y[ii]
        c_los_z = ldeti[ii]['cents_z'] + k_los * los_z[ii]

        # get projections of corners on plane perp. to los
        sca0 = (
            (c_los_x - det_Px[None, :]) * los_x[ii]
            + (c_los_y - det_Py[None, :]) * los_y[ii]
            + (c_los_z - det_Pz[None, :]) * los_z[ii]
        )
        k_plane = sca0 / sca1

        # get LOS-specific unit vectors

        e0_xi = (
            los_y[ii] * ldeti[ii]['e1_z'] - los_z[ii] * ldeti[ii]['e1_y']
        )
        e0_yi = (
            los_z[ii] * ldeti[ii]['e1_x'] - los_x[ii] * ldeti[ii]['e1_z']
        )
        e0_zi = (
            los_x[ii] * ldeti[ii]['e1_y'] - los_y[ii] * ldeti[ii]['e1_x']
        )

        e0_normi = np.sqrt(e0_xi**2 + e0_yi**2 + e0_zi**2)
        e0_xi = e0_xi / e0_normi
        e0_yi = e0_yi / e0_normi
        e0_zi = e0_zi / e0_normi

        e1_xi = los_y[ii] * e0_zi - los_z[ii] * e0_yi
        e1_yi = los_z[ii] * e0_xi - los_x[ii] * e0_zi
        e1_zi = los_x[ii] * e0_yi - los_y[ii] * e0_xi

        # get projections on det_e0 and det_e1 in plane

        x0 = np.split(
            ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e0_xi
            + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e0_yi
            + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e0_zi,
            ap_ind,
            axis=0,
        )
        x1 = np.split(
            ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e1_xi
            + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e1_yi
            + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e1_zi,
            ap_ind,
            axis=0,
        )

        x0_min = np.max([np.min(x0s) for x0s in x0])
        x0_max = np.min([np.max(x0s) for x0s in x0])
        x1_min = np.max([np.min(x1s) for x1s in x1])
        x1_max = np.min([np.max(x1s) for x1s in x1])

        w0 = x0_max - x0_min
        w1 = x1_max - x1_min

        min_res = min(2*margin_perp*w0, 2*margin_perp*w1)
        too_large = res >= min_res
        if np.any(too_large):
            msg = (
                f"Minimum etendue resolution for det {ii} / {nd}: {min_res}\n"
                "The following res values may lead to errors:\n"
                f"\t- res values = {res}\n"
                f"\t- too large  = {too_large}\n"
            )
            warnings.warn(msg)

        # -------------------
        # Discretize aperture

        for jj in range(res.size):

            coef = 1. + 2.*margin_perp
            n0 = int(np.ceil(coef*w0 / res[jj]))
            n1 = int(np.ceil(coef*w1 / res[jj]))

            d0 = coef*w0 / n0
            d1 = coef*w1 / n1

            ds = d0 * d1

            pts_0 = np.linspace(
                x0_min - margin_perp*w0,
                x0_max + margin_perp*w0,
                n0 + 1,
            )
            pts_1 = np.linspace(
                x1_min - margin_perp*w1,
                x1_max + margin_perp*w1,
                n1 + 1,
            )
            pts_0 = 0.5 * (pts_0[1:] + pts_0[:-1])
            pts_1 = 0.5 * (pts_1[1:] + pts_1[:-1])

            # debug
            # n0, n1 = 2, 2
            # pts_0 = np.r_[pts_0[0], pts_0[0]]
            # pts_1 = np.r_[0, 0]

            pts_x = (
                c_los_x + pts_0[:, None] * e0_xi + pts_1[None, :] * e1_xi
            ).ravel()
            pts_y = (
                c_los_y + pts_0[:, None] * e0_yi + pts_1[None, :] * e1_yi
            ).ravel()
            pts_z = (
                c_los_z + pts_0[:, None] * e0_zi + pts_1[None, :] * e1_zi
            ).ravel()

            if verb is True:
                msg = (
                    f"\tres = {res[jj]} ({jj+1} / {res.size})"
                    f"    nb. of points = {pts_x.size}"
                )
                print(msg)

            # ----------------------------------
            # compute solid angle for each pixel

            if check is True:
                solid_angle = _comp_solidangles.calc_solidangle_apertures(
                    # observation points
                    pts_x=pts_x,
                    pts_y=pts_y,
                    pts_z=pts_z,
                    # polygons
                    apertures=aperture,
                    detectors=ldeti[ii],
                    # possible obstacles
                    config=None,
                    # parameters
                    visibility=False,
                    return_vector=False,
                    return_flat_pts=True,
                    return_flat_det=True,
                )

                sar = solid_angle.reshape((n0, n1))
                c0 = (
                    ((pts_0[0] < x0_min) == np.all(sar[0, :] == 0))
                    and ((pts_0[-1] > x0_max) == np.all(sar[-1, :] == 0))
                    and ((pts_1[0] < x1_min) == np.all(sar[:, 0] == 0))
                    and ((pts_1[-1] > x1_max) == np.all(sar[:, -1] == 0))
                )
                if not c0 and not too_large[jj]:
                    # debug
                    plt.figure()
                    plt.imshow(
                        sar.T,
                        extent=(
                            x0_min - margin_perp*w0, x0_max + margin_perp*w0,
                            x1_min - margin_perp*w1, x1_max + margin_perp*w1,
                        ),
                        interpolation='nearest',
                        origin='lower',
                        aspect='equal',
                    )

                    lc = ['r', 'm', 'c', 'y']
                    for ss in range(len(x0)):
                        iss = np.r_[np.arange(0, x0[ss].shape[0]), 0]
                        plt.plot(
                            x0[ss][iss, :],
                            x1[ss][iss, :],
                            c=lc[ss%len(lc)],
                            marker='o',
                            ls='-',
                        )

                    plt.plot(
                        pts_0, np.mean(pts_1)*np.ones((n0,)),
                        c='k', marker='.', ls='None',
                    )
                    plt.plot(
                        np.mean(pts_0)*np.ones((n1,)), pts_1,
                        c='k', marker='.', ls='None',
                    )
                    plt.gca().set_xlabel('x0')
                    plt.gca().set_xlabel('x1')
                    # import pdb; pdb.set_trace()
                    msg = "Something is wrong with solid_angle or sampling"
                    raise Exception(msg)
                else:
                    etendue[jj, ii] = np.sum(solid_angle) * ds

            else:
                etendue[jj, ii] = _comp_solidangles.calc_solidangle_apertures(
                    # observation points
                    pts_x=pts_x,
                    pts_y=pts_y,
                    pts_z=pts_z,
                    # polygons
                    apertures=aperture,
                    detectors=ldeti[ii],
                    # possible obstacles
                    config=None,
                    # parameters
                    summed=True,
                    visibility=False,
                    return_vector=False,
                    return_flat_pts=True,
                    return_flat_det=True,
                ) * ds

    return etendue


# #############################################################################
# #############################################################################
#                   Plotting routine
# #############################################################################


def _plot_etendues(
    etend0=None,
    etend1=None,
    res=None,
):

    # -------------
    # prepare data

    nmax = 0
    if etend0 is not None:
        if etend0.ndim > 2:
            etend0 = etend0.reshape((etend0.shape[0], -1))
        nmax = max(nmax, etend0.shape[0])
    if etend1 is not None:
        if etend1.ndim > 2:
            etend1 = etend1.reshape((etend1.shape[0], -1))
        nmax = max(nmax, etend1.shape[0])

    x0 = None
    if etend0 is not None:
        x0 = [
            f'order {ii}' if ii < 3 else '' for ii in range(nmax)
        ]
    if etend1 is not None:
        x1 = [f'{res[ii]}' if ii < res.size-1 else '' for ii in range(nmax)]
        if x0 is None:
            x0 = x1
        else:
            x0 = [f'{x0[ii]}\n{x1[ii]}' for ii in range(nmax)]

    # -------------
    # prepare axes

    fig = plt.figure(figsize=(11, 6))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    ax.set_ylabel('Etendue ' + r'($m^2.sr$)', size=12, fontweight='bold')
    ax.set_xlabel('order of approximation', size=12, fontweight='bold')

    ax.set_xticks(range(0, nmax))
    ax.set_xticklabels(x0)

    # -------------
    # plot

    if etend0 is not None:
        lines = ax.plot(
            etend0,
            ls='-',
            marker='o',
            ms=6,
        )
        lcol = [ll.get_color() for ll in lines]
    else:
        lcol = [None for ii in range(etend1.shape[1])]

    if etend1 is not None:
        for ii in range(etend1.shape[1]):
            ax.plot(
                etend1[:, ii],
                ls='--',
                marker='*',
                ms=6,
                color=lcol[ii],
            )

    # -------------
    # legend

    handles = [
        mlines.Line2D(
            [], [],
            c='k', marker='o', ls='-', ms=6,
            label='analytical',
        ),
        mlines.Line2D(
            [], [],
            c='k', marker='*', ls='--', ms=6,
            label='numerical',
        ),
    ]
    ax.legend(handles=handles)

    return ax



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True
#
from libc.stdlib cimport malloc, free


cdef void insert(SortedSet** head, double val) nogil:
  # insert the value at the right position so all elements
  # are in increasing order, don't add if already present
  cdef SortedSet* adding
  cdef SortedSet* current
  cdef SortedSet* previous

  # set is empty
  if head[0] == NULL:
      adding = <SortedSet*>malloc(sizeof(SortedSet))
      adding.val = val
      adding.next = NULL
      adding.last = adding
      head[0] = adding
      return

  # adding on top
  if head[0].val > val:
      adding = <SortedSet*>malloc(sizeof(SortedSet))
      adding.val = val
      adding.next = head[0]
      adding.last = head[0].last
      head[0] = adding
      return

  # adding on bottom
  if head[0].last.val < val:
      adding = <SortedSet*>malloc(sizeof(SortedSet))
      adding.val = val
      adding.next = NULL
      head[0].last.next = adding
      head[0].last = adding
      return

  # if same as one of extremities, do nothing
  if head[0].val == val or head[0].last.val == val:
      return

  current = head[0].next
  previous = head[0]
  while current.val < val:
      previous = current
      current = current.next

  # already in set, nothing to do
  if current.val == val:
      return

  adding = <SortedSet*>malloc(sizeof(SortedSet))
  adding.val = val
  adding.next = current
  previous.next = adding

  return


cdef int count(SortedSet* head, double val) nogil:
  # returns 1 if val in set, else 0
  cdef SortedSet* current

  # outside of scope
  if head.val > val or head.last.val < val:
      return 0

  # is one of extremities
  if  head.val == val or head.last.val == val:
      return 1

  current = head[0].next
  while current.val <= val:
      if current.val == val:
          return 1
      current = current.next

  # we ended and didnt find a match
  return 0

cdef void free_ss(SortedSet** head) nogil:
  cdef SortedSet* current

  if head[0] == NULL :
      return

  previous = head[0]
  current = head[0].next
  while current != NULL:
      free(previous)
      previous = current
      current = current.next
  free(previous)
  return




# Common
import numpy as np
import scipy.signal as scpsig
import scipy.interpolate as scpinterp

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from tofu.version import __version__


__all__ = [
    'get_localextrema_1d',
    'peak_analysis_spect1d',
    'plot_peak_analysis_spect1d',
]


_GITHUB = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = f'tofu-{__version__}\treport issues / requests at {_GITHUB}'


# Useful scalar types
_NINT = (np.int32, np.int64)
_INT = (int,) + _NINT
_NFLOAT = (np.float32, np.float64)
_FLOAT = (float,) + _NFLOAT
_NUMB = _INT + _FLOAT


###########################################################
###########################################################
#           hidden utilities
###########################################################


def _check_bool(var, vardef=None, varname=None):
    if var is None:
        var = vardef
    if not isinstance(var, bool):
        msg = (
            "Arg {} must be a bool\n".format(varname)
            + "  You provided: {}".format(type(var))
        )
        raise Exception(msg)
    return var


###########################################################
###########################################################
#
#           Preliminary
#       utility tools for 1d spectral analysis
#
###########################################################
###########################################################


def _get_localextrema_1d_check(
    data=None, lamb=None,
    weights=None, width=None,
    prom_rel=None, rel_height=None,
    method=None, returnas=None,
    return_minima=None,
    return_prominence=None,
    return_width=None,
):
    # data
    c0 = (
        isinstance(data, np.ndarray)
        and data.ndim in [1, 2]
        and data.size > 0
        and np.all(np.isfinite(data))
    )
    if not c0:
        msg = (
            "Arg data must be a (nlamb,) or (nt, nlamb) finite np.array!\n"
            + "\t- provided: {}\n".format(data)
        )
        raise Exception(msg)
    if data.ndim == 1:
        data = data[None, :]

    # lamb
    if lamb is None:
        lamb = np.arange(data.shape[1])
    c0 = (
        isinstance(lamb, np.ndarray)
        and lamb.shape == (data.shape[1],)
        and np.all(np.isfinite(lamb))
        and np.all(np.diff(lamb) > 0)
    )
    if not c0:
        msg = (
            "Arg lamb must be a finite increasing (data.shape[0],) np.array!\n"
            + "\t- provided: {}".format(lamb)
        )
        raise Exception(msg)

    # method
    if method is None:
        method = 'find_peaks'
    c0 = method in ['find_peaks', 'bspline']
    if not c0:
        msg = (
            "Arg method must be either:\n"
            + "\t- 'find_peaks': uses scipy.signal.find_peaks()\n"
            + "\t- 'bspline': uses bspline-fitting to find extrema\n"
            + "You provided:\n{}".format(method)
        )
        raise Exception(msg)

    # weights (for fitting, optional)
    c0 = (
        weights is None
        or (
            isinstance(weights, np.ndarray)
            and weights.shape == (data.shape[1],)
        )
    )
    if not c0:
        msg = (
            "Arg weights must be either None or a (nlamb,) np.array!\n"
            + "Fed to scipy.interpolate.UnivariateSpline(w=weights)\n"
            + "\t- provided: {}".format(weights)
        )
        raise Exception(msg)

    # width
    if width is None:
        if method == 'find_peaks':
            width = 0.
        else:
            width = False
    c0 = width is False or (isinstance(width, _NUMB) and width >= 0.)
    if not c0:
        msg = (
            "Arg width must be a float\n"
            + "width = estimate of the minimum line width\n"
            + "If method == 'find_peaks':\n"
            + "=>  used as minimum peak width\n"
            + "If method == 'bspline':\n"
            + "=>  Used to smooth the bsplines fitting\n"
            + "\t- False: spline fits all points\n"
            + "\t- float > 0: spline is smoothed\n"
            + "You provided: {}".format(width)
        )
        raise Exception(msg)
    if width is not False and method == 'find_peaks':
        width = int(np.ceil(width / np.nanmean(np.diff(lamb))))

    if rel_height is None:
        rel_height = 0.8
    if not (isinstance(rel_height, _NUMB) and 0 <= rel_height <= 1.):
        msg = (
            "Arg rel_height must be positive float in [0, 1]!\n"
            + "Provided: {}".format(rel_height)
        )
        raise Exception(msg)

    # returnas
    if returnas is None:
        returnas = float
    c0 = returnas in [bool, float]
    if not c0:
        msg = (
            "Arg returnas must be:\n"
            + "\t- bool: return 2 (nt, nlamb) bool arrays, True at extrema\n"
            + "\t- float: return 2 (nt, nn) float arrays, of extrema values\n"
            + "  You provided:\n{}".format(returnas)
        )
        raise Exception(msg)

    # return_minima, prominence, width
    return_minima = _check_bool(
        return_minima, vardef=False, varname='return_minima',
    )
    return_prominence = _check_bool(
        return_prominence, vardef=True, varname='return_prominence',
    )
    return_width = _check_bool(
        return_width, vardef=True, varname='return_width',
    )
    prom_rel = _check_bool(
        prom_rel, vardef=True, varname='prom_rel',
    )

    return (
        data, lamb, weights, width, prom_rel, rel_height, method,
        returnas, return_minima, return_prominence, return_width,
    )


def get_localextrema_1d(
    data=None, lamb=None,
    width=None, weights=None,
    prom_rel=None, rel_height=None,
    distance=None,
    method=None, returnas=None,
    return_minima=None,
    return_prominence=None,
    return_width=None,
):
    """ Automatically find peaks in spectrum """

    # ------------------
    #   check inputs
    (
        data, lamb, weights,
        width, prom_rel, rel_height, method,
        returnas, return_minima,
        return_prominence, return_width
    ) = _get_localextrema_1d_check(
        data=data, lamb=lamb,
        weights=weights, width=width,
        prom_rel=prom_rel, rel_height=rel_height,
        method=method, returnas=returnas,
        return_minima=return_minima,
        return_prominence=return_prominence,
        return_width=return_width,
    )

    # -----------------
    #   fit and extract extrema
    mini, prom, widt = None, None, None
    minima, prominence, widths = None, None, None
    if method == 'find_peaks':

        # find_peaks
        maxi = np.zeros(data.shape, dtype=bool)
        if return_prominence is True:
            prom = np.full(data.shape, np.nan)
        if return_width is True:
            widt = np.full(data.shape, np.nan)

        for ii in range(data.shape[0]):
            peaks, prop = scpsig.find_peaks(
                data[ii, :], height=None, threshold=None,
                distance=distance, prominence=0,
                width=width, wlen=None, rel_height=rel_height,
                plateau_size=None,
            )
            maxi[ii, peaks] = True
            if return_prominence is True:
                prom[ii, peaks] = prop['prominences']
            if return_width is True:
                widt[ii, peaks] = prop['widths']

        if return_minima is True:
            mini = np.zeros(data.shape, dtype=bool)
            for ii in range(data.shape[0]):
                peaks, prop = scpsig.find_peaks(
                    -data[ii, :], height=None, threshold=None,
                    distance=distance, prominence=None,
                    width=width, wlen=None, rel_height=rel_height,
                    plateau_size=None,
                )
                mini[ii, peaks] = True

        #   reshape
        if returnas is float:
            nmax = np.sum(maxi, axis=1)
            if return_prominence is True:
                prominence = np.full((data.shape[0], np.max(nmax)), np.nan)
                for ii in range(data.shape[0]):
                    prominence[ii, :nmax[ii]] = prom[ii, maxi[ii, :]]
            if return_width is True:
                widths = np.full((data.shape[0], np.max(nmax)), np.nan)
                for ii in range(data.shape[0]):
                    widths[ii, :nmax[ii]] = widt[ii, maxi[ii, :]]

            maxima = np.full((data.shape[0], np.max(nmax)), np.nan)
            for ii in range(data.shape[0]):
                maxima[ii, :nmax[ii]] = lamb[maxi[ii, :]]
            if return_minima is True:
                nmax = np.sum(mini, axis=1)
                minima = np.full((data.shape[0], np.max(nmax)), np.nan)
                for ii in range(data.shape[0]):
                    minima[ii, :nmax[ii]] = lamb[mini[ii, :]]

        else:
            maxima = maxi
            minima = mini
            prominence = prom
            widths = widt

        if prominence is not None and prom_rel is True:
            prominence = prominence / np.nanmax(data, axis=1)[:, None]

    else:

        # bspline
        prominence, widths = None, None
        bbox = [lamb.min(), lamb.max()]
        mini, maxi = [], []
        if width is False:
            for ii in range(data.shape[0]):
                bs = scpinterp.UnivariateSpline(
                    lamb, data[ii, :], w=weights,
                    bbox=bbox, k=4,
                    s=0, ext=2,
                    check_finite=False,
                )
                extrema = bs.derivative(1).roots()
                indmin = bs.derivative(2)(extrema) > 0.
                indmax = bs.derivative(2)(extrema) < 0.
                mini.append(extrema[indmin])
                maxi.append(extrema[indmax])
        else:
            nint = int(np.ceil((bbox[1]-bbox[0]) / (1.1*width)))
            delta = (bbox[1]-bbox[0]) / nint
            nknots = nint - 1
            knots = np.linspace(bbox[0]+delta, bbox[1]-delta, nknots)
            for ii in range(data.shape[0]):
                bs = scpinterp.LSQUnivariateSpline(
                    lamb, data[ii, :], t=knots,
                    w=weights,
                    bbox=bbox, k=4, ext=2,
                    check_finite=False,
                )
                extrema = bs.derivative(1).roots()
                indmin = bs.derivative(2)(extrema) > 0.
                indmax = bs.derivative(2)(extrema) < 0.
                mini.append(extrema[indmin])
                maxi.append(extrema[indmax])

        #   reshape
        if returnas is bool:
            bins = 0.5*(lamb[1:] + lamb[:-1])
            bins = np.r_[
                bins[0]-(lamb[1]-lamb[0]),
                bins,
                bins[-1]+(lamb[-1]-lamb[-2]),
            ]
            minima = np.zeros(data.shape, dtype=bool)
            maxima = np.zeros(data.shape, dtype=bool)
            for ii in range(data.shape[0]):
                if len(mini[ii]) > 0:
                    minima[
                        ii,
                        np.digitize(mini[ii], bins, right=False)-1
                    ] = True
                if len(maxi[ii]) > 0:
                    maxima[
                        ii,
                        np.digitize(maxi[ii], bins, right=False)-1
                    ] = True
        else:
            nmin = np.max([len(mm) for mm in mini])
            nmax = np.max([len(mm) for mm in maxi])
            minima = np.full((data.shape[0], nmin), np.nan)
            maxima = np.full((data.shape[0], nmax), np.nan)
            for ii in range(data.shape[0]):
                minima[ii, :len(mini[ii])] = mini[ii]
                maxima[ii, :len(maxi[ii])] = maxi[ii]

    # ---------------
    # return
    lout = [
        (minima, return_minima),
        (prominence, return_prominence),
        (widths, return_width),
    ]
    out = [maxima] + [aa for aa, vv in lout if vv is True]
    if len(out) == 1:
        out = out[0]
    return out


def peak_analysis_spect1d(
    intervals=None,
    data=None, lamb=None, t=None,
    thresh_prom=None,
    thresh_time=None,
    thresh_time_higher=None,
    thresh_lines_prom=None,
    groupby=None,
    distance=None,
    width=None, weights=None,
    prom_rel=None, rel_height=None,
    method=None, returnas=None,
    return_minima=None,
    return_prominence=None,
    return_width=None,
):

    # ------------------
    #   check inputs
    (
        data, lamb, weights,
        width, prom_rel, rel_height, method,
        returnas, return_minima,
        return_prominence, return_width
    ) = _get_localextrema_1d_check(
        data=data, lamb=lamb,
        weights=weights, width=width,
        prom_rel=prom_rel, rel_height=rel_height,
        method=method, returnas=returnas,
        return_minima=return_minima,
        return_prominence=return_prominence,
        return_width=return_width,
    )

    # time
    if t is None:
        t = np.arange(0, data.shape[0])

    # intervals
    if intervals is None:
        intervals = [(-np.inf, np.inf)]

    # thresholds
    if thresh_prom is None:
        thresh_prom = 0.
    if thresh_time is None:
        thresh_time = 0.
    if thresh_time_higher is None:
        thresh_time_higher = True
    if thresh_lines_prom is None:
        thresh_lines_prom = 0.
    if groupby is None:
        groupby = False

    # Get peaks and prominences
    maxi, prom = get_localextrema_1d(
        data=data,
        lamb=lamb,
        distance=distance,
        prom_rel=True,
        returnas=bool,
        return_minima=False,
        return_prominence=True,
        return_width=False,
    )

    # by-interval
    prom0 = np.copy(prom)
    dt = np.mean(np.diff(t))

    prom0[prom0 < thresh_prom] = np.nan
    # if thresh_prom is not None:
    # for ii, (i0, i1) in enumerate(intervals):
    # ind = (t >= i0) & (t <= i1)
    # prom0[ind[:, None] & np.all(prom0[ind, :]< thresh_prom, axis=0)] = np.nan

    if thresh_time is not None:
        fract_time = (~np.isnan(prom0))
        for ii, (i0, i1) in enumerate(intervals):
            deltai = (intervals[ii][1]-intervals[ii][0])/dt
            ind = (t >= i0) & (t <= i1)
            indi = np.sum(fract_time[ind, :], axis=0) > thresh_time*ind.sum()
            if thresh_time_higher is False:
                indi = ~indi
            fract_time[ind, :] = fract_time[ind, :] & indi[None, :]
        prom0[~fract_time] = np.nan

    hist = np.zeros((len(intervals), data.shape[1]), dtype=float)
    for ii, (i0, i1) in enumerate(intervals):
        ind = (t >= i0) & (t <= i1)
        hist[ii, :] = np.nansum(prom0[ind, :], axis=0)

    if groupby is not False:
        hist0 = np.copy(hist)
        for ii in range(hist0.shape[0]):
            cont = True
            ic = 0
            while cont:
                inds = np.argsort(hist0[ii, :])[::-1]
                for kk, jj in enumerate(inds):
                    i0 = np.arange(max(0, jj-groupby), jj)
                    i1 = np.arange(jj+1, min(jj+groupby+1, hist0.shape[1]))
                    c0 = (
                        hist0[ii, jj] > 0.
                        and (
                            np.any(hist0[ii, i0] > 0.)
                            or np.any(hist0[ii, i1] > 0.)
                        )
                    )
                    if c0:
                        hist0[ii, i0] = 0.
                        hist0[ii, i1] = 0.
                        ic += 1
                        break
                    else:
                        pass
                    if jj == inds[-1]:
                        cont = False
        lines = [
            lamb[hist0[ii, :] > thresh_lines_prom]
            for ii in range(hist0.shape[0])
        ]
    else:
        lines = [
            lamb[hist[ii, :] > thresh_lines_prom]
            for ii in range(hist.shape[0])
        ]

    # Return
    danalysis = {
        'data': data, 'lamb': lamb, 't': t,
        'maxi': maxi, 'prom': prom,
        'intervals': intervals, 'hist': hist, 'lines': lines,
        'thresh_prom': thresh_prom,
        'thresh_time': thresh_time,
        'thresh_lines_prom': thresh_lines_prom,
    }
    return danalysis


def plot_peak_analysis_spect1d(
    danalysis,
    sharey=None,
    dax=None,
    cmap=None, vmin=None, vmax=None,
    fs=None, dmargin=None,
    tit=None, wintit=None,
):

    # Prepare data
    # ------------
    nint = len(danalysis['intervals'])

    # Check plot inputs
    # ------------------

    if fs is None:
        fs = (14, 8)
    if tit is None:
        tit = False
    if wintit is None:
        wintit = _WINTIT
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.07, 'top': 0.92,
                   'wspace': 0.2, 'hspace': 0.3}
    if sharey is None:
        sharey = len(danalysis['lines']) > 1

    if dax is None:
        fig = plt.figure(figsize=fs)
        if tit is not False:
            fig.suptitle(tit, size=14, fontweight='bold')

        gs = gridspec.GridSpec(1, 2, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0])
        sharey = ax0 if sharey else None
        ax1 = fig.add_subplot(gs[0, 1], sharex=ax0, sharey=sharey)

        ax0.set_xlabel(r'$\lambda$')
        ax0.set_ylabel(r'$t$')
        ax1.set_xlabel(r'$\lambda$')
        ax0.set_title('Peaks prominence', size=14, fontweight='bold')
        ax1.set_title('Peaks per interval', size=14, fontweight='bold')

        dax = {
            'peaks': ax0,
            'hist': ax1,
        }

    k0 = 'peaks'
    if dax.get(k0) is not None:
        prom0 = np.copy(danalysis['prom'])
        prom0[np.isnan(prom0)] = 0
        extent = (
            danalysis['lamb'].min(), danalysis['lamb'].max(),
            danalysis['t'].min(), danalysis['t'].max(),
        )
        dax[k0].imshow(
            prom0,
            vmin=vmin, vmax=vmax,
            cmap=cmap,
            extent=extent,
            interpolation='nearest',
            origin='lower',
            aspect='auto',
        )

    k0 = 'hist'
    if dax.get(k0) is not None:
        for ii, (i0, i1) in enumerate(danalysis['intervals']):
            indt = (danalysis['t'] >= i0) & (danalysis['t'] <= i1)
            if sharey is not None:
                intmin = np.min(danalysis['t'][indt])
                intdelta = (
                    danalysis['t'][indt].max() - danalysis['t'][indt].min()
                )
                ymax = np.nanmax(danalysis['hist'][ii, :])
                y = danalysis['hist'][ii, :] * (intdelta/ymax) + intmin
                ythr = (
                    danalysis['thresh_lines_prom'] * (intdelta/ymax) + intmin
                )
            else:
                intmin = 0.
                y = danalysis['hist'][ii, :]
                ythr = danalysis['thresh_lines_prom']
            dax[k0].plot(danalysis['lamb'], y, c='k', ls='-')
            dax[k0].axhline(intmin, c='k', ls='--')
            dax[k0].axhline(ythr, c='k', ls='--')

    return dax




import datetime as dtm
import itertools as itt


import numpy as np
import scipy.integrate as scpinteg
import astropy.units as asunits


import datastock as ds


# ################################################################
# ################################################################
#               Main routine
# ################################################################


def compute_signal(
    coll=None,
    key=None,
    key_diag=None,
    key_cam=None,
    # to be integrated
    key_integrand=None,
    # spectral ref
    key_ref_spectro=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    groupby=None,
    val_init=None,
    ref_com=None,
    # vos
    dvos=None,
    # signal
    brightness=None,
    spectral_binning=None,
    # verb
    verb=None,
    # timing
    timing=None,
    # store
    store=None,
    # return
    returnas=None,
):

    # -------------
    # check inputs
    # --------------

    (
        key_diag, key_cam, spectro, PHA, is2d,
        method, mode, groupby, val_init,
        brightness, spectral_binning,
        key_integrand, key_mesh0, key_bs,
        key_ref_spectro, key_bs_spectro,
        dvos,
        verb, timing, store, key,
        returnas,
    ) = _compute_signal_check(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        # sampling
        method=method,
        res=res,
        mode=mode,
        groupby=groupby,
        val_init=val_init,
        # vos
        dvos=dvos,
        # signal
        brightness=brightness,
        spectral_binning=spectral_binning,
        # to be integrated
        key_integrand=key_integrand,
        # spectral ref
        key_ref_spectro=key_ref_spectro,
        # verb
        verb=verb,
        # timing
        timing=timing,
        # store
        store=store,
        key=key,
        # return
        returnas=returnas,
    )

    # -------------
    # prepare
    # --------------

    # shape_emiss = coll.ddata[key_integrand]['shape']

    if mode == 'abs':
        key_kR = coll.dobj['mesh'][key_mesh0]['knots'][0]
        radius_max = np.max(coll.ddata[key_kR]['data'])
    else:
        radius_max = None

    # -------------
    # verb
    # --------------

    if verb is True:
        msg = (
            "\nComputing synthetic signal for:\n"
            f"\t- diag: {key_diag}\n"
            f"\t- cam: {key_cam}\n"
            f"\t- integrand: {key_integrand}\n"
            f"\t- method: {method}\n"
            f"\t- res: {res}, {mode}\n"
        )
        print(msg)

    # -------------
    # compute
    # --------------

    # pick routine
    if method == 'los':
        func = _compute_los
    else:
        if spectro is True:
            func = _compute_vos_spectro
        else:
            func = _compute_vos_broadband

    # call routine
    dout, dt = func(
        coll=coll,
        is2d=is2d,
        spectro=spectro,
        PHA=PHA,
        key=key,
        key_diag=key_diag,
        key_cam=key_cam,
        key_bs=key_bs,
        res=res,
        mode=mode,
        key_integrand=key_integrand,
        key_ref_spectro=key_ref_spectro,
        key_bs_spectro=key_bs_spectro,
        radius_max=radius_max,
        groupby=groupby,
        val_init=val_init,
        ref_com=ref_com,
        brightness=brightness,
        spectral_binning=spectral_binning,
        # vos
        dvos=dvos,
        # verb
        verb=verb,
        # timing
        timing=timing,
    )

    # ----------
    # timing
    # ----------

    if timing is True:
        lstr = [f"{k0}:\t {v0} s" for k0, v0 in dt.items()]
        msg = (
            "Timing for "
            f"compute_diagnostic_signal('{key_diag}', method='{method}')\n"
            + "\n".join(lstr)
        )
        print(msg)

    # -------------
    # store
    # --------------

    if store is True:
        _store(
            coll=coll,
            key=key,
            key_diag=key_diag,
            dout=dout,
            key_integrand=key_integrand,
            method=method,
            res=res,
        )

    # -------------
    # return
    # --------------

    if returnas is dict:
        return dout


# ##################################################################
# ##################################################################
#               STORE
# ##################################################################


def _store(
    coll=None,
    key=None,
    key_diag=None,
    dout=None,
    # synthetic signal
    key_integrand=None,
    method=None,
    res=None,
    # retrofit
    key_matrix=None,
):

    # ---------
    # check

    lc = [key_integrand is not None, key_matrix is not None]
    if np.sum(lc) != 1:
        msg = "Please provide key_integrand xor key_matrix"
        raise Exception(msg)

    # ---------
    # prepare

    lkcam = list(dout.keys())
    dobj = {
        'synth sig':{
            key: {
                'diag': key_diag,
                'camera': lkcam,
                'data': [dout[k0]['key'] for k0 in lkcam],
                # synthetic
                'integrand': key_integrand,
                'method': method,
                'res': res,
                # retrofit
                'geom matrix': key_matrix,
            },
        },
    }

    # ----------
    # add data

    for k0, v0 in dout.items():
        coll.add_data(**v0)

    # add obj
    lsynth = coll._dobj['diagnostic'][key_diag].get('signal')
    if lsynth is None:
        lsynth = [key]
    else:
        lsynth.append(key)
    coll._dobj['diagnostic'][key_diag]['signal'] = lsynth
    coll.update(dobj=dobj)


# ##################################################################
#               CHECK
# ##################################################################


def _compute_signal_check(
    coll=None,
    key=None,
    key_diag=None,
    key_cam=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    groupby=None,
    val_init=None,
    # vos
    dvos=None,
    # signal
    brightness=None,
    spectral_binning=None,
    # to be integrated
    key_integrand=None,
    # spectral ref
    key_ref_spectro=None,
    # verb
    verb=None,
    # timing
    timing=None,
    # store
    store=None,
    # return
    returnas=None,
):

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    # key_diag, key_cam
    key_diag, key_cam = coll.get_diagnostic_cam(key=key_diag, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    PHA = coll.dobj['diagnostic'][key_diag]['PHA']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # method
    method = ds._generic_check._check_var(
        method, 'method',
        types=str,
        default='los',
        allowed=['los', 'vos'],
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='abs',
        allowed=['abs', 'rel'],
    )

    # groupby
    groupby = ds._generic_check._check_var(
        groupby, 'groupby',
        types=int,
        default=1 if (PHA or spectro) else 200,
        allowed=[1] if (PHA or spectro) else None,
    )

    # brightness
    brightness = ds._generic_check._check_var(
        brightness, 'brightness',
        types=bool,
        default=False,
    )

    # key_integrand
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0.get(wbs) is not None
        and any([
            coll.dobj[wm][coll.dobj[wbs][k1][wm]]['nd'] == '2d'
            or coll.dobj[wm][coll.dobj[wbs][k1][wm]]['submesh'] is not None
            for k1 in v0[wbs]
        ])
    ]

    key_integrand = ds._generic_check._check_var(
        key_integrand, 'key_integrand',
        types=str,
        allowed=lok,
    )

    # key_mesh0
    key_bs = [
        kk for kk in coll.ddata[key_integrand][wbs]
        if coll.dobj[wm][coll.dobj[wbs][kk][wm]]['nd'] == '2d'
        or coll.dobj[wm][coll.dobj[wbs][kk][wm]]['submesh'] is not None
    ]
    if len(key_bs) == 1:
        key_bs = key_bs[0]
    else:
        msg = f"Multiple possible 2d bsplines for integrand '{key_integrand}'"
        raise Exception(msg)

    key_mesh = coll.dobj[wbs][key_bs][wm]
    submesh = coll.dobj[wm][key_mesh]['submesh']
    if submesh is not None:
        key_mesh0 = submesh
    else:
        key_mesh0 = key_mesh

    # key_ref_spectro
    if spectro:
        key_ref_spectro, key_bs_spectro = _get_ref_bs_spectro(
            coll=coll,
            key_integrand=key_integrand,
            key_bs=key_bs,
            key_ref_spectro=key_ref_spectro,
        )
    else:
        key_ref_spectro = None
        key_bs_spectro = None

    # val_init
    val_init = ds._generic_check._check_var(
        val_init, 'val_init',
        default=np.nan,
        allowed=[np.nan, 0.]
    )

    # dvos
    if method == 'vos':
        # single camera + get dvos
        key_diag, dvos, isstore = coll.check_diagnostic_dvos(
            key_diag,
            key_cam=key_cam,
            dvos=dvos,
        )

        if spectro is True and (not isstore):
            msg = "spectro synthetic signal with 'vos' required stored vos!"
            raise Exception(msg)

    # verb
    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # timing
    timing = ds._generic_check._check_var(
        timing, 'timing',
        types=bool,
        default=False,
    )

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # key
    if store is True:
        key = ds._generic_check._obj_key(
            coll.dobj.get('synth sig', {}),
            short='synth',
            key=key,
            ndigits=2,
        )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=False if store is True else dict,
        allowed=[dict, False],
    )

    return (
        key_diag, key_cam, spectro, PHA, is2d,
        method, mode, groupby, val_init,
        brightness, spectral_binning,
        key_integrand, key_mesh0, key_bs,
        key_ref_spectro, key_bs_spectro,
        dvos,
        verb, timing, store, key,
        returnas,
    )


def _get_ref_bs_spectro(
    coll=None,
    key_integrand=None,
    key_bs=None,
    key_ref_spectro=None,
):

    # -----------
    # prepare

    # get ref of integrand
    kref = coll.ddata[key_integrand]['ref']
    wbs = coll._which_bsplines

    # get list of non-spatial ref
    lkspectro = [
        k0 for k0 in kref
        if k0 not in coll.dobj[wbs][key_bs]['ref']
    ]

    # If none => error
    if len(lkspectro) == 0:
        msg = (
            f"Integrand '{key_integrand}' does not seem to "
            "have a spectral dimension"
        )
        raise Exception(msg)

    # lbs_spectro
    lbs_spectro = [
        k0 for k0 in coll.ddata[key_integrand][wbs]
        if k0 != key_bs
    ]

    # --------------------------
    # Determine key_ref_spectro

    key_bs_spectro = None
    if key_ref_spectro is None:

        # unique => ok
        if len(lkspectro) == 1:
            key_ref_spectro = lkspectro[0]
        else:
            pass

    # check if bs
    if len(lbs_spectro) == 0:
        pass
    elif len(lbs_spectro) == 1:
        key_bs_spectro = lbs_spectro[0]
        if key_ref_spectro is None:
            key_ref_spectro = coll.dobj[wbs][key_bs_spectro]['ref'][0]
        assert key_ref_spectro == coll.dobj[wbs][key_bs_spectro]['ref'][0]
    else:
        pass

    # --------
    # safety check

    if key_ref_spectro is None:
        msg = f"Spectral dimension of '{key_integrand} could not be identified'"
        raise Exception(msg)

    return key_ref_spectro, key_bs_spectro


# ################################################################
# ################################################################
#               LOS
# ################################################################


def _compute_los(
    coll=None,
    is2d=None,
    spectro=None,
    PHA=None,
    key=None,
    key_diag=None,
    key_cam=None,
    key_bs=None,
    res=None,
    mode=None,
    key_integrand=None,
    key_ref_spectro=None,
    key_bs_spectro=None,
    radius_max=None,
    groupby=None,
    val_init=None,
    ref_com=None,
    brightness=None,
    spectral_binning=None,
    # verb
    verb=None,
    # timing
    timing=None,
    # unused
    **kwdargs,
):

    # --------------
    # prepare timing

    dt = None
    if timing is True:
        lk = [
            '\tpreparation',
            '\tsample rays',
            '\tspectral binning',
            '\tinterpolate on los',
            '\textract data',
            '\tintegrate on los',
            '\tformat output'
        ]
        dt = dict.fromkeys(lk, 0)
        t0 = dtm.datetime.now()     # Timing

    # -----------------
    # prepare

    if spectro:

        kspect_ref_vect = coll.get_ref_vector(ref=key_ref_spectro)[3]
        spect_ref_vect = coll.ddata[kspect_ref_vect]['data']

        ref = coll.ddata[key_integrand]['ref']
        axis_spectro = ref.index(key_ref_spectro)

        wbs = coll._which_bsplines
        rbs = coll.dobj[wbs][key_bs]['ref'][0]
        if axis_spectro > ref.index(rbs):
            axis_spectro -= len(coll.dobj[wbs][key_bs]['ref']) - 1

        units_spectro = coll.ddata[kspect_ref_vect]['units']

        # --------------------------
        # optional spectral binning

        # spectral_binning
        spectral_binning = ds._generic_check._check_var(
            spectral_binning, 'spectral_binning',
            types=bool,
            default=True,
        )

        # if spectral binning => add bins of len 2 for temporary storing
        if spectral_binning is True:
            ktemp_bin = f'{key_ref_spectro}_temp_bin'
            coll.add_bins(
                key=ktemp_bin,
                edges=[0, 1],
                units=units_spectro,
            )
            ktemp_binc = coll.dobj['bins'][ktemp_bin]['cents'][0]

    else:
        pass

    # units
    units0, units_bs = _units_integration(
        coll=coll,
        key_integrand=key_integrand,
        key_bs=key_bs,
    )
    units = units0 * units_bs
    domain = None

    # timing
    if timing is True:
        t1 = dtm.datetime.now()
        dt['\tpreparation'] = (t1 - t0).total_seconds()

    # ----------------
    # loop on cameras

    key_integrand_interp = str(key_integrand)
    dout = {}
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    for k0 in key_cam:

        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        key_los = doptics[k0]['los']
        key_pts0 = coll.dobj['rays'][key_los]['pts'][0]
        ilosok = np.isfinite(coll.ddata[key_pts0]['data'][0, ...].ravel())

        ngroup = npix // groupby
        if groupby * ngroup < npix:
            ngroup += 1

        # ----------------
        # spectro

        if spectro:
            E, _ = coll.get_diagnostic_lamb(
                key_diag,
                lamb='lamb',
                key_cam=k0,
                units=units_spectro,
            )
            dE, _ = coll.get_diagnostic_lamb(
                key_diag,
                lamb='dlamb',
                key_cam=k0,
                units=units_spectro,
            )
            E_flat = E.ravel()
            dE_flat = dE.ravel()

        # ---------------------------------------------------
        # loop on group of pixels (to limit memory footprint)

        shape_cam = coll.dobj['camera'][k0]['dgeom']['shape']
        shape = None
        for ii in range(ngroup):

            # verb
            if verb is True:
                msg = f"\tpix group {ii+1} / {ngroup}"
                end = "\n" if ii == ngroup - 1 else "\r"
                print(msg, end=end, flush=True)

            # timing
            if timing is True:
                t01 = dtm.datetime.now()

            # indices
            i0 = ii*groupby
            i1 = min((ii + 1)*groupby, npix)
            ni = i1 - i0

            # get rid of undefined LOS
            ind_ch_flat = np.array([jj for jj in range(i0, i1) if ilosok[jj]], dtype=int)
            ind_ch = np.unravel_index(ind_ch_flat, shape_cam)
            ni = ind_ch_flat.size

            # no valid los in group
            if ni == 0:
                continue

            # LOS sampling
            R, Z, length = coll.sample_rays(
                key=key_los,
                res=res,
                mode=mode,
                segment=None,
                ind_ch=ind_ch,
                radius_max=radius_max,
                concatenate=True,
                return_coords=['R', 'z', 'ltot'],
            )

            if R is None:
                continue

            # safety checks
            inan = np.isnan(R)
            inannb = np.r_[-1, inan.nonzero()[0]]
            nnan = inan.sum()

            # some lines can be nan if non-existant
            # one nan for each separation between concatenated LOS
            if nnan != ni:
                msg = (
                    "Mismatch between expected and observed nb of NaNs!\n"
                    f"\t- key_diag = {key_diag}\n"
                    f"\t- kcam = {k0}\n"
                    f"\t- ii = {ii}\n"
                    f"\t- i1 - i0 = {i1} - {i0} = {ni}\n"
                    f"\t- groupby, npix = {groupby}    {npix}\n"
                    f"\t- ni = ind_ch_flat.size = {ni}\n"
                    f"\t- shape_cam = {shape_cam}\n"
                    f"\t- ind_ch = {ind_ch}\n"
                    f"\t- res, mode = {res}   {mode}\n"
                    f"\t- R.shape = {R.shape}\n"
                    f"\t- nnan = {nnan}\n"
                )
                raise Exception(msg)

            # timing
            if timing is True:
                t02 = dtm.datetime.now()
                dt['\tsample rays'] += (t02-t01).total_seconds()

            # -------------------
            # domain for spectro

            if spectro and spectral_binning:

                # add bins for storing
                coll._ddata[ktemp_binc]['data'] = E_flat[ii]
                edges = E_flat[ii] + dE_flat[ii] * 0.5 * np.r_[-1, 1]
                coll._dobj['bins'][ktemp_bin]['edges'] = edges

                # bin spectrally before spatial interpolation
                kbinned = f"{key_integrand}_bin_{k0}_{ii}"
                #try:
                coll.binning(
                    data=key_integrand,
                    bin_data0=key_ref_spectro if key_bs_spectro is None else key_bs_spectro,
                    bins0=ktemp_bin,
                    integrate=True,
                    verb=verb,
                    store=True,
                    returnas=False,
                    store_keys=kbinned,
                )
                # except Exception as err:
                #     msg = (
                #         err.args[0]
                #         + "\n\n"
                #         f"\t- k0 = {k0}\n"
                #         f"\t- ii = {ii}\n"
                #         f"\t- key_integrand = {key_integrand}\n"
                #         f"\t- key_bs_spectro = {key_bs_spectro}\n"
                #         f"\t- ktemp_bin = {ktemp_bin}\n"
                #         f"\t- edges = {edges}\n"
                #         f"\t- E_flat[ii] = {E_flat[ii]}\n"
                #     )
                #     err.args = (msg,)
                #     raise err

                domain = None
                key_integrand_interp = kbinned

            elif spectro:
                ind = np.argmin(np.abs(spect_ref_vect - E_flat[ind_ch_flat[0]]))
                domain = {key_ref_spectro: {'ind': np.r_[ind]}}

            # timing
            if timing is True:
                t03 = dtm.datetime.now()
                dt['\tspectral binning'] += (t03-t02).total_seconds()

            # ---------------------
            # interpolate spacially

            # datai, units, refi = coll.interpolate(
            douti = coll.interpolate(
                keys=key_integrand_interp,
                ref_key=key_bs,
                x0=R,
                x1=Z,
                grid=False,
                submesh=True,
                ref_com=ref_com,
                domain=domain,
                # azone=None,
                details=False,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )[key_integrand_interp]

            # timing
            if timing is True:
                t04 = dtm.datetime.now()
                dt['\tinterpolate on los'] += (t04-t03).total_seconds()

            # ----------------------
            # interpolate spectrally

            if spectro:
                douti['data'] = np.take(douti['data'], 0, axis_spectro)

                if spectral_binning is True:
                    coll.remove_data(kbinned)

                douti['ref'] = tuple([
                    rr for jj, rr in enumerate(douti['ref'])
                    if jj != axis_spectro
                ])

            # ------------
            # extract data

            datai, refi = douti['data'], douti['ref']
            axis = refi.index(None)
            if shape is None:
                shape = list(datai.shape)
                shape[axis] = npix
                data = np.full(shape, val_init)
                ref = list(refi)

            # timing
            if timing is True:
                t05 = dtm.datetime.now()
                dt['\textract data'] += (t05-t04).total_seconds()

            # ------------
            # integrate

            iok2 = np.isfinite(datai)
            sli0 = [slice(None) for aa in range(len(refi))]
            for jj in range(nnan):

                # slice datai
                indi = np.arange(inannb[jj]+1, inannb[jj+1])
                sli0[axis] = indi
                slii = tuple(sli0)
                if not np.any(iok2[slii]):
                    continue

                # set nan to 0 for integration
                dataii = datai[slii]
                dataii[~iok2[slii]] = 0.

                # slice data
                sli0[axis] = ind_ch_flat[jj]
                sli = tuple(sli0)

                # if jj in [50, 51]:
                    # plt.figure();
                    # plt.subplot(1,2,1)
                    # plt.plot(dataii)
                    # plt.subplot(1,2,2)
                    # plt.plot(dataii.T)
                    # plt.gcf().suptitle(f"jj = {jj}", size=12)

                # integrate
                data[sli] = scpinteg.trapezoid(
                    dataii,
                    x=length[indi],
                    axis=axis,
                )

            # timing
            if timing is True:
                t06 = dtm.datetime.now()
                dt['\tintegrate on los'] += (t06-t05).total_seconds()

        # --------------
        # safety check

        if shape is None:
            msg = (
                "Looks like no single LOS can see non-zero emissivity!\n"
                f"\t- key_los = {key_los}\n"
                f"\t- ni = {ni}\n"
                f"\t- R = {R}\n"
                "\ncoll.show('rays'):\n"
                + coll.show('rays', returnas=str, verb=False)
            )
            raise Exception(msg)

        # --------------
        # post-treatment

        # brightness
        if brightness is False:
            ketend = doptics[k0]['etendue']
            etend = coll.ddata[ketend]['data']
            sh_etend = [-1 if aa == axis else 1 for aa in range(len(refi))]
            data *= etend.reshape(sh_etend)
            unitsi = units * coll.ddata[ketend]['units']
        else:
            unitsi = units

        # spectral bins if spectro
        if spectro:
            unitsi = unitsi * units_spectro
            if spectral_binning is True:
                pass
            else:
                sh_dE = [-1 if aa == axis else 1 for aa in range(len(refi))]
                data *= dE_flat.reshape(sh_dE)

        # reshape if 2d
        if is2d:
            sh_data = list(data.shape)
            sh_data[axis] = coll.dobj['camera'][k0]['dgeom']['shape']
            sh_data = tuple(np.r_[
                sh_data[:axis], sh_data[axis], sh_data[axis+1:]
            ].astype(int))
            data = data.reshape(sh_data)

        # set ref
        ref[axis] = coll.dobj['camera'][k0]['dgeom']['ref']
        ref = tuple(np.r_[ref[:axis], ref[axis], ref[axis+1:]])

        # timing
        if timing is True:
            t07 = dtm.datetime.now()
            dt['\tformat output'] += (t07-t06).total_seconds()

        # fill dout
        dout[k0] = {
            'key': f'{key}_{k0}',
            'data': data,
            'ref': ref,
            'units': unitsi,
        }

    # ----------
    # clean up

    if spectro is True and spectral_binning is True:
        coll.remove_bins(ktemp_bin)

    return dout, dt


def _units_integration(
    coll=None,
    key_integrand=None,
    key_bs=None,
):

    # -----------------
    # integrand

    # units0
    units0 = coll.ddata[key_integrand]['units']

    # --------------
    # los dimension

    # units_bs
    wbs = coll._which_bsplines
    kap = coll.dobj[wbs][key_bs]['apex']
    lunits = list({coll.ddata[k0]['units'] for k0 in kap})
    if len(lunits) == 1:
        units_bs = lunits[0]
    else:
        msg = "Don't know how to interpret line-integration units from bspline"
        raise Exception(msg)

    # units_bs from subbs if relevant
    wm = coll._which_mesh
    keym = coll.dobj[wbs][key_bs][wm]
    subbs = coll.dobj[wm][keym]['subbs']
    if subbs is not None:
        kap = coll.dobj[wbs][subbs]['apex']
        lunits = list({coll.ddata[k0]['units'] for k0 in kap})
        if len(lunits) == 1:
            units_bs = lunits[0]
        else:
            msg = "Don't know how to interpret line-integration units from bspline"
            raise Exception(msg)

    return units0, units_bs


# ##################################################################
# ##################################################################
#               VOS - Broadband
# ##################################################################


def _compute_vos_broadband(
    coll=None,
    is2d=None,
    PHA=None,
    key=None,
    key_diag=None,
    key_cam=None,
    key_bs=None,
    res=None,
    mode=None,
    key_integrand=None,
    key_ref_spectro=None,
    key_bs_spectro=None,
    key_ref_cos=None,
    groupby=None,
    val_init=None,
    ref_com=None,
    brightness=None,
    spectral_binning=None,
    # dvos
    dvos=None,
    # verb
    verb=None,
    # unused
    **kwdargs,
):

    # ------------------------
    # check uniformity of dvos

    dt = None
    # check all keym and res_RZ are similar
    lkm = list(set([v0['keym'] for v0 in dvos.values()]))
    lres = list(set([tuple(v0['res_RZ']) for v0 in dvos.values()]))

    if len(lkm) != 1:
        lstr = [f"\t- '{k0}': '{v0['keym']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same mesh!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    if len(lres) != 1:
        lstr = [f"\t- '{k0}': '{v0['res_RZ']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same resolution!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # extract
    keym = lkm[0]
    res_RZ = list(lres[0])

    # -----------------
    # get mesh sampling

    # get dsamp
    dsamp = coll.get_sample_mesh(
        key=keym,
        res=res_RZ,
        mode='abs',
        grid=False,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    x0u = dsamp['x0']['data']
    x1u = dsamp['x1']['data']

    # --------------
    # prepare

    # units
    units0 = coll.ddata[key_integrand]['units']

    # -----------------------------
    # loop on cameras

    dout = {k0: {} for k0 in dvos.keys()}
    for k0, v0 in dvos.items():

        # group
        units_vos = v0['sang_cross']['units']

        # --------------
        # loop on pixels

        shape = None
        shape_cam = v0['indr_cross']['data'].shape[:-1]
        assert shape_cam == coll.dobj['camera'][k0]['dgeom']['shape']
        ref_cam = coll.dobj['camera'][k0]['dgeom']['ref']
        for ind in np.ndindex(shape_cam):

            # no valid los in group
            iok = v0['indr_cross']['data'][ind] >= 0
            if not np.any(iok):
                continue

            # vos re-creation
            ind_RZ = tuple(list(ind) + [iok])
            R = x0u[v0['indr_cross']['data'][ind_RZ]]
            Z = x1u[v0['indz_cross']['data'][ind_RZ]]

            # -----------------------------
            # interpolate on matching wavelength ?

            douti = coll.interpolate(
                keys=key_integrand,
                ref_key=key_bs,
                x0=R,
                x1=Z,
                grid=False,
                submesh=True,
                ref_com=ref_com,
                domain=None,
                # azone=None,
                details=False,
                crop=None,
                nan0=False,
                val_out=0.,
                return_params=False,
                store=False,
            )[key_integrand]

            # extracti shape and ref from integrand
            datai, refi = douti['data'], douti['ref']
            if shape is None:
                axis = refi.index(None)
                shape = list(datai.shape)
                shape = tuple(
                    np.r_[shape[:axis], shape_cam, shape[axis+1:]].astype(int)
                )
                ref = tuple(np.r_[refi[:axis], ref_cam, refi[axis+1:]])
                data = np.full(shape, val_init)
                ind_data = [[slice(None)] for ii in range(datai.ndim)]
                ind_sa = [[None] for ii in range(datai.ndim)]

            # ------------
            # integrate

            ind_data[axis] = ind
            ind_sa[axis] = ind_RZ
            data[tuple(itt.chain.from_iterable(ind_data))] = np.nansum(
                datai
                * v0['sang_cross']['data'][tuple(itt.chain.from_iterable(ind_sa))],
                axis=axis,
            )

        # --------------
        # post-treatment

        unitsi = units0 * units_vos

        # fill dout
        dout[k0] = {
            'key': f'{key}_{k0}',
            'data': data,
            'ref': ref,
            'units': unitsi,
        }

    # ----------
    # clean up

    return dout, dt


# ##################################################################
# ##################################################################
#               VOS - spectro
# ##################################################################


def _compute_vos_spectro(
    coll=None,
    is2d=None,
    spectro=None,
    PHA=None,
    key=None,
    key_diag=None,
    key_cam=None,
    key_bs=None,
    res=None,
    mode=None,
    key_integrand=None,
    key_ref_spectro=None,
    key_bs_spectro=None,
    key_ref_cos=None,
    radius_max=None,
    groupby=None,
    val_init=None,
    ref_com=None,
    brightness=None,
    spectral_binning=None,
    # dvos
    dvos=None,
    # verb
    verb=None,
    # unused
    **kwdargs,
):
    """

    spectral_binning = bool
        if True, bins spectrum
        if False, interpolates spectrum

    """

    # --------------
    # prepare timing

    dt = None

    # ------------------------
    # check uniformity of dvos

    # check all keym and res_RZ are similar
    lkm = list(set([v0['keym'] for v0 in dvos.values()]))
    lres = list(set([tuple(v0['res_RZ']) for v0 in dvos.values()]))

    if len(lkm) != 1:
        lstr = [f"\t- '{k0}': '{v0['keym']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same mesh!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    if len(lres) != 1:
        lstr = [f"\t- '{k0}': '{v0['res_RZ']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same resolution!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # extract
    keym = lkm[0]
    res_RZ = list(lres[0])

    # -----------------
    # get mesh sampling

    # get dsamp
    dsamp = coll.get_sample_mesh(
        key=keym,
        res=res_RZ,
        mode='abs',
        grid=False,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    x0u = dsamp['x0']['data']
    x1u = dsamp['x1']['data']

    # --------------
    # prepare

    kspect_ref_vect = coll.get_ref_vector(ref=key_ref_spectro)[3]
    units_spectro = coll.ddata[kspect_ref_vect]['units']

    wbs = coll._which_bsplines

    # -----------------------------
    # interpolate per wavelength ?

    dout = {k0: {} for k0 in dvos.keys()}
    for k0, v0 in dvos.items():

        R = x0u[v0['indr_cross']['data']]
        Z = x1u[v0['indz_cross']['data']]

        kapex = coll.dobj[wbs][key_bs_spectro]['apex'][0]
        dlamb_ref = np.mean(np.diff(coll.ddata[kapex]['data']))
        lamb = v0['lamb']['data']
        dlamb = np.mean(np.diff(lamb))
        if spectral_binning is None:
            spectral_binningi = dlamb > dlamb_ref
        else:
            spectral_binningi = spectral_binning

        # ------------------------------------------------------
        # interpolate on matching wavelength or bin spectrally?

        # pre-interpolate at lamb
        # (nt, nbs0, nbs1, nlamb_ref, ncos) => (nt, nbs0, nbs1, nlamb, ncos)
        # or
        # (nt, nbs0, nbs1, nlamb_ref) => (nt, nbs0, nbs1, nlamb)

        key_integrand_interp_lamb = f"{key_integrand}_interp_lamb"
        if spectral_binningi is True:

            ktemp_bin = f'{key_bs_spectro}_{k0}_temp_bin'
            edge_spectro = np.r_[
                lamb[0] - 0.5*(lamb[1] - lamb[0]),
                0.5 * (lamb[1:] + lamb[:-1]),
                lamb[-1] + 0.5*(lamb[-1] - lamb[-2]),
            ]

            coll.add_bins(
                key=ktemp_bin,
                edges=edge_spectro,
                units=units_spectro,
            )

            coll.binning(
                data=key_integrand,
                bin_data0=key_bs_spectro,
                bins0=ktemp_bin,
                integrate=True,
                verb=verb,
                store=True,
                returnas=False,
                store_keys=key_integrand_interp_lamb,
            )

            # adjust ref
            refb = coll.dobj['bins'][ktemp_bin]['ref'][0]
            iref = coll.ddata[key_integrand_interp_lamb]['ref'].index(refb)
            # safety check
            shapei = coll.ddata[key_integrand_interp_lamb]['data'].shape
            assert shapei[iref] == lamb.size
            refb = list(coll.ddata[key_integrand_interp_lamb]['ref'])
            refb[iref] = v0['lamb']['ref'][0]
            coll.ddata[key_integrand_interp_lamb]['ref'] = tuple(refb)

            # remove bins
            coll.remove_bins(ktemp_bin)

        else:

            _ = coll.interpolate(
                keys=key_integrand,
                ref_key=key_bs_spectro,
                x0=v0['lamb']['key'],
                x1=None,
                grid=False,
                submesh=None,
                ref_com=None,
                domain=None,
                # azone=None,
                details=False,
                crop=None,
                nan0=False,
                val_out=0.,
                return_params=False,
                store=True,
                inplace=True,
                store_keys=key_integrand_interp_lamb,
            )

            # adjust data and units for integration
            dlamb = np.mean(np.diff(v0['lamb']['data']))
            coll.ddata[key_integrand_interp_lamb]['data'] *= dlamb
            coll.ddata[key_integrand_interp_lamb]['units'] = (
                asunits.Unit(coll.ddata[key_integrand_interp_lamb]['units'])
                * asunits.Unit(v0['lamb']['units'])
            )

        # -----------------------------
        # interpolate vs local cos

        if key_ref_cos is not None:

            # (nt, nbs0, nbs1, nlamb, ncos) => (nt, nbs0, nbs1, nlamb)
            douti = coll.interpolate(
                keys=key_integrand_interp_lamb,
                ref_key=key_bs,
                x0=R,
                x1=Z,
                grid=False,
                submesh=True,
                ref_com=ref_com,
                domain=None,
                # azone=None,
                details=False,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )[key_integrand_interp_lamb]

        # -----------------------------
        # interpolate per spatial pts

        # (nt, nbs0, nbs1, nlamb) => (nt, npts, nlamb)

        douti = coll.interpolate(
            keys=key_integrand_interp_lamb,
            ref_key=key_bs,
            x0=R,
            x1=Z,
            grid=False,
            submesh=True,
            ref_com=ref_com,
            domain=None,
            # azone=None,
            details=False,
            crop=None,
            nan0=True,
            val_out=np.nan,
            return_params=False,
            store=False,
        )[key_integrand_interp_lamb]

        # -------------------
        # sum to get signal

        # ref
        ref_vos = list(coll.ddata[v0['ph']['key']]['ref'])
        ref_data = list(douti['ref'])
        ref_data[ref_data.index(None)] = ref_vos[-2]
        refc = [rr for rr in ref_data if rr in ref_vos]
        ref = (
            ref_data[:ref_data.index(refc[0])]
            + ref_vos[:ref_vos.index(refc[0])]
            + refc
        )

        # (nt, npts, nlamb)
        shape_data = douti['data'].shape

        # (n0, n1, npts, nlamb)
        shape_vos = v0['ph']['data'].shape

        # (nt, n0, n1)
        shape_sig = [
            shape_data[ref_data.index(rr)]
            if rr in ref_data else shape_vos[ref_vos.index(rr)]
            for ii, rr in enumerate(ref)
            if rr not in refc
        ]

        # slicing
        sli_vos = [
            slice(None) if rr in ref_vos[-2:]
            else (0 if rr in ref_vos else None)
            for ii, rr in enumerate(ref)
        ]

        # slicing sig
        sli_sig = [
            slice(None) if rr not in ref_vos else 0
            for ii, rr in enumerate(ref)
            if rr not in ref_vos[-2:]
        ]

        # axis for summation
        ax_lamb_data = ref_data.index(ref_vos[-1])
        ax_pts_data = ref_data.index(ref_vos[-2])
        ax_n0_vos = ref.index(ref_vos[0])
        ax_n1_vos = ref.index(ref_vos[1])
        ax_n0_sig = ref[:-2].index(ref_vos[0])
        ax_n1_sig = ref[:-2].index(ref_vos[1])

        # sum on pts and lamb
        # loop to save memory
        sig = np.full(shape_sig, np.nan)
        for (i0, i1) in itt.product(range(shape_vos[0]), range(shape_vos[1])):
            sli_sig[ax_n0_sig] = i0
            sli_sig[ax_n1_sig] = i1
            sli_vos[ax_n0_vos] = i0
            sli_vos[ax_n1_vos] = i1
            sig[tuple(sli_sig)] = np.nansum(
                douti['data']
                * v0['ph']['data'][tuple(sli_vos)],
                axis=(ax_lamb_data, ax_pts_data),
            )

        # new ref
        ref = [rr for rr in ref if rr not in refc]

        # units
        units = asunits.Unit(v0['ph']['units']) * asunits.Unit(douti['units'])

        # --------------
        # post-treatment

        # brightness
        if brightness is True:
            ketend = coll.dobj['diagnostic'][key_diag]['doptics'][k0]['etendue']
            etend = coll.ddata[ketend]['data']
            ref_e = coll.ddata[ketend]['ref']
            sh_etend = (
                [1]*ref.index(ref_e[0])
                + ref_e
                + [1]*(len(ref) - ref.index(ref_e[-1]) - 1)
            )
            sig /= etend.reshape(sh_etend)
            units = units / coll.ddata[ketend]['units']
        else:
            units = units

        # --------------
        # storing

        dout[k0] = {
            'key': f'{key}_{k0}',
            'data': sig,
            'ref': ref,
            'units': units,
        }

        # ----------
        # clean up

        if spectral_binning is True:
            coll.remove_bins(ktemp_bin)
        coll.remove_data(key_integrand_interp_lamb)

    return dout, dt


# -*- coding: utf-8 -*-


# Built-in
# import copy


# tofu
from ._class04_Filter import Filter as Previous
from . import _class3_check
from . import _class5_check as _check
from . import _class5_compute as _compute
from . import _class05_legacy as _legacy
from . import _class5_coordinates as _coordinates
from . import _class5_reflections_pts2pt as _reflections_pts2pt
from . import _class5_reflections_ptsvect as _reflections_ptsvect
from . import _class5_projections as _projections
from . import _class5_plot as _plot


__all__ = ['Crystal']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Crystal(Previous):

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'crystal': [
            'dgeom.type',
            'dgeom.curve_r',
            'dgeom.area',
            'dmat.name',
            'dmat.miller',
            'dgeom.outline',
            'dgeom.poly',
            'dgeom.cent',
            'dmisc.color',
        ],
    })

    def add_crystal(
        self,
        key=None,
        # geometry
        dgeom=None,
        # material
        dmat=None,
        alpha=None,
        beta=None,
        # dmisc
        color=None,
    ):
        """ Add a crystal

        Can be defined from:
            - 2d outline + 3d center + unit vectors (nin, e0, e1)
            - 3d polygon + nin

        Unit vectors will be checked and normalized
        If planar, area will be computed
        Outline will be made counter-clockwise

        """

        # check / format input
        dref, ddata, dobj = _class3_check._add_surface3d(
            coll=self,
            key=key,
            which='crystal',
            which_short='cryst',
            # 2d outline
            **dgeom,
        )

        key = list(dobj['crystal'].keys())[0]

        # material
        dobj['crystal'][key]['dmat'] = _check._dmat(
            coll=self,
            key=key,
            dgeom=dobj['crystal'][key]['dgeom'],
            dmat=dmat,
            alpha=alpha,
            beta=beta,
        )

        # dmisc
        dobj['crystal'][key]['dmisc'] = _class3_check._dmisc(
            key=key,
            color=color,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # compute rocking curve
        # if dmat['ready_to_compute'] is True:
        # self.set_crystal_rocking_curve()

    def set_crystal_rocking_curve(
        self,
        key=None,
        dbragg=None,
        intensity=None,
    ):
        return _compute.rocking_curve(
            coll=self,
            key=key,
        )

    # --------------------
    # Legacy utility
    # ---------------------

    def add_crystal_from_legacy(
        self,
        cryst=None,
        key=None,
    ):
        """ Add a crystal from a tf.geom.CrystalBragg instance (or pfe) """
        return _legacy.add_crystal(
            self,
            cryst=cryst,
            key=key,
        )

    # --------------------
    # Utilities
    # ---------------------

    def get_crystal_bragglamb(
        self,
        key=None,
        lamb=None,
        bragg=None,
        norder=None,
        rocking_curve=None,
    ):
        """ Return bragg angle

        If bragg provided, simply return bragg as np.ndarray

        If lamb is provided return corresponding:
            - bragg angle (simple bragg's law)
            - bragg angle + reflectivity, interpolated on rocking_curve

        """

        return _compute._bragglamb(
            coll=self,
            key=key,
            lamb=lamb,
            bragg=bragg,
            rocking_curve=rocking_curve,
        )

    # --------------------
    # Local vs xyz coordinates
    # ---------------------

    def get_optics_x01toxyz(
        self,
        key=None,
        asplane=None,
    ):
        """ Return a dict of formatted """
        return _coordinates._get_x01toxyz(
            coll=self,
            key=key,
            asplane=asplane,
        )

    # --------------------
    # Reflections
    # ---------------------

    def get_optics_reflect_pts2pt(self, key=None):
        """ Return a dict of formatted """
        return _reflections_pts2pt._get_pts2pt(
            coll=self,
            key=key,
        )

    def get_optics_reflect_ptsvect(
        self,
        key=None,
        asplane=None,
        isnorm=None,
        fast=None,
    ):
        """ Return a dict of formatted """
        return _reflections_ptsvect._get_ptsvect(
            coll=self,
            key=key,
            asplane=asplane,
            isnorm=isnorm,
            fast=fast,
        )

    # -------------------------------------------
    # Projections from point in local coordinates
    # -------------------------------------------

    def get_optics_project_poly_from_pt(self, key=None):
        """ Return a dict of formatted """
        return _projections._get_pts_from_pt(coll=self, key=key)

    # --------------------
    # ideal configurations
    # ---------------------

    def get_crystal_ideal_configuration(
        self,
        key=None,
        configuration=None,
        lamb=None,
        bragg=None,
        defocus=None,
        strict_aperture=None,
        # parameters
        cam_on_e0=None,
        # johann-specific
        cam_tangential=None,
        # pinhole-specific
        cam_dimensions=None,
        cam_distance=None,
        focal_distance=None,
        # store
        store=None,
        key_cam=None,
        key_aperture=None,
        aperture_dimensions=None,
        pinhole_radius=None,
        cam_pixels_nb=None,
        # returnas
        returnas=None,
    ):
        """ Return the ideal positions of other elements for a configuration

        'ideal' means:
            - maximizing focalization
            - for the chosen wavelength / bragg angle

        'johann': for spherical crystals only
            build the rowland circle and return the ideal detector position

        'von hamos': for cylindrical only
            build the axis and return
             - slit position
             - detector position given its height

        'pinhole': for cylindrical only
            buids the axis and return
             - pinhole position
             - detector position given its height
        """

        return _compute._ideal_configuration(
            coll=self,
            key=key,
            configuration=configuration,
            lamb=lamb,
            bragg=bragg,
            defocus=defocus,
            strict_aperture=strict_aperture,
            # parameters
            cam_on_e0=cam_on_e0,
            # johann-specific
            cam_tangential=cam_tangential,
            # pinhole-specific
            cam_dimensions=cam_dimensions,
            cam_distance=cam_distance,
            focal_distance=focal_distance,
            # store
            store=store,
            key_cam=key_cam,
            key_aperture=key_aperture,
            aperture_dimensions=aperture_dimensions,
            pinhole_radius=pinhole_radius,
            cam_pixels_nb=cam_pixels_nb,
            # returnas
            returnas=returnas,
        )

    # --------------------
    # plotting
    # ---------------------

    def plot_crystal_rocking_curve(
        self,
        key=None,
        # option
        T=None,
        # plotting
        dax=None,
        color=None,
        plot_FW=None,
    ):

        return _plot.plot_rocking_curve(
            coll=self,
            key=key,
            # option
            T=T,
            # plotting
            dax=dax,
            color=color,
            plot_FW=plot_FW,
        )




import numpy as np
import scipy.constants as scpct


# ##############################################################
# ##############################################################
#           Conversions momentum - velocity - energy
# ##############################################################


def convert_momentum_velocity_energy(
    energy_kinetic_eV=None,
    velocity_ms=None,
    momentum_normalized=None,
    gamma=None,
    beta=None,
):
    """ Convert any input to all outputs

    returns a dict with, for each ouput key 'data' and 'units'

    - momentum_normalized: total relativistic momentum / mec2
    - energy_kinetic_eV: kinetic energy in eV
    - gamma: Lorentz factor
    - beta = v / c
    - velocity_ms : velocity in m/s

    see:
    [1] Pandya et al., Physica Scripta 93, no. 11 (November 1, 2018): 115601
    [2] https://en.wikipedia.org/wiki/Energy%E2%80%93momentum_relation

    """

    # ---------------
    # dict in - check
    # ---------------

    din0 = locals()
    din = {k0: v0 for k0, v0 in din0.items() if v0 is not None}

    if len(din) != 1:
        lstr = [f"\t- {k0}" for k0 in din.keys()]
        msg = (
            "Please provide only one input of the following:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    key = list(din.keys())[0]
    val = np.atleast_1d(din[key])

    # -----------------
    # convert to gamma
    # -----------------

    if 'gamma' not in din.keys():
        gamma = _to_gamma(key=key, val=val)

    else:
        gamma = val

    # -----------------
    # initialize dout
    # -----------------

    dout = {
        'gamma': {
            'data': gamma,
            'units': None,
        },
    }

    # -----------------
    # convert from gamma
    # -----------------

    lk = [k0 for k0 in din0.keys() if k0 != 'gamma']
    for k0 in lk:
        dout[k0] = _from_gamma(key=k0, gamma=gamma)

    return dout


def _to_gamma(key, val):

    if key == 'beta':
        gamma = np.sqrt(1. / (1. - val**2))

    elif key == 'velocity_ms':
        gamma = np.sqrt(1. / (1. - (val/scpct.c)**2))

    elif key == 'momentum_normalized':
        gamma = np.sqrt(val**2 + 1)

    elif key == 'energy_kinetic_eV':
        mc2_eV = scpct.m_e * scpct.c**2 / scpct.e
        gamma = (val + mc2_eV) / mc2_eV

    else:
        msg = f"key {key} not implemented in _to_gamma()!"
        raise Exception(msg)

    return gamma


def _from_gamma(key, gamma):

    if key == 'beta':
        out = np.sqrt(gamma**2 - 1) / gamma
        units = None

    elif key == 'velocity_ms':
        out = scpct.c * np.sqrt(gamma**2 - 1) / gamma
        units = 'm/s'

    elif key == 'momentum_normalized':
        out = np.sqrt(gamma**2 - 1)
        units = None

    elif key == 'energy_kinetic_eV':
        mc2_eV = scpct.m_e * scpct.c**2 / scpct.e
        out = mc2_eV * (gamma - 1)
        units = 'eV'

    else:
        msg = f"key {key} not implemented in _from_gamma()!"
        raise Exception(msg)

    return {'data': out, 'units': units}






from . import benchmarks_00_Geometry_peakmem
from . import benchmarks_00_Geometry_time
from . import benchmarks_01_Mesh2D_peakmem
from . import benchmarks_01_Mesh2D_time
from . import benchmarks_02_spectralfit_peakmem
from . import benchmarks_02_spectralfit_time
from . import benchmarks_03_solidangles



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    TEST magFieldLines
'''
from __future__ import (unicode_literals, absolute_import,  \
                        print_function, division)
import doctest
import os
import sys

if __name__ == '__main__':
    #print('path 1 =', sys.path)
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
    #print('path 2 =', sys.path)
    #import mag
    from mag import magFieldLines
    path = sys.path.pop(0)
    doctest.testmod(magFieldLines, verbose=True)
